/*
 * $Id: bakers.c,v 1.5 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"

#define BakersDeck	0
#define BakersPile	(BakersDeck + 1)
#define BakersNpile	4
#define BakersLpile	(BakersPile + BakersNpile - 1)
#define BakersTable	(BakersLpile + 1)
#define BakersNtable	13
#define BakersLtable	(BakersTable + BakersNtable - 1)
#define BakersNstack	(BakersLtable + 1)

CardSetupType	BakersSetup[] = {
    { { 0, 0 }, { 0, 0 }, CardEmpty, CardDisplayNone, False, False, 0, CardStartDown },

    // pile / foundation
    { { 7, 0 }, { 7, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 7, 1 }, { 7, 1 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 7, 2 }, { 7, 2 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 7, 3 }, { 7, 3 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },

    // table
    { { 0, 0 }, { 0, 0 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 1, 0 }, { 1, 0 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 2, 0 }, { 2, 0 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 3, 0 }, { 3, 0 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 4, 0 }, { 4, 0 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 5, 0 }, { 5, 0 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 6, 0 }, { 6, 0 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 0, 3 }, { 0, 3 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 1, 3 }, { 1, 3 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 2, 3 }, { 2, 3 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 3, 3 }, { 3, 3 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 4, 3 }, { 4, 3 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },
    { { 5, 3 }, { 5, 3 }, CardEmpty, CardDisplayAll, False, False, 4, CardStartUp },

};

static void BakersInitialize (void) GAME_SECT;
static CardBool BakersSelectTableUnordered (CardStackIndex stack,
        CardIndex *first, CardIndex *last) GAME_SECT;
static CardProgressType BakersCheckTable (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType BakersCheckPile (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardBool BakersNormalize (void) GAME_SECT;

static void
BakersInitialize (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    CardIndex	    card;
    CardType	    *cp;
    CardIndex	    stop_card, next;

    for (stack = BakersTable; stack <= BakersLtable; stack++)
    {
        sp = CardStackPtr (stack);
        // Move Kings to bottom of stack
        card = sp->last;
        stop_card = sp->first;
        while (card != stop_card)
	{
	    cp = CardPtr (card);
            next = cp->prev;
            if (Rank (cp->card) == KING)
            {
                // move king to front
                CardMoveCard (stack, card, card, stack, NullIndex, False);
            }
            card = next;
        }
    }
}

// Only allow moving one card, and only if it's not already on
// an in-order card
static CardBool
BakersSelectTableUnordered (CardStackIndex   stack,
		   CardIndex	    *first,
		   CardIndex	    *last)
{
    CardStackType   *sp = CardStackPtr (stack);
    CardIndex	    card;
    CardType	    *cp;

    card = sp->last;
    *first = card;
    *last = card;
    cp = CardPtr (card);
    if (cp->prev != NullIndex && CardIsInOrder (card, cp->prev))
        return False;
    return True;
}

static CardProgressType
BakersCheckTable (CardStackIndex	src, 
		 CardStackIndex dst,
		 CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (CardIsInOrder (card, dp->last))
	return CardProgressSome;
    return CardProgressCant;
}

static CardProgressType
BakersCheckPile (CardStackIndex	src, 
		CardStackIndex	dst,
		CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
    {
	if (Rank (CardPtr (card)->card) == 1)
	    return CardProgressSome;
    }
    else
    {
	if (CardIsInSuitOrder (dp->last, card))
	    return CardProgressSome;
    }
    return CardProgressCant;
}


CardMoveType	BakersMoves[] = {
    /* 0    Move card from table to pile */
    {	"table to pile",
	{ BakersTable, BakersLtable },
	{ BakersTable, BakersLtable },
	{ BakersPile, BakersLpile },
	CardMoveAuto,
	0,
	BakersCheckPile,
	0,
    },
    /* 1    Move card from table to table */
    {	"table to table (if not ordered)",
	{ BakersTable, BakersLtable },
	{ BakersTable, BakersLtable },
	{ BakersTable, BakersLtable },
	CardMoveHelpful,
	BakersSelectTableUnordered,
	BakersCheckTable,
	0,
    },
    /* 2    Move card from table to table */
    {	"table to table",
	{ BakersTable, BakersLtable },
	{ BakersTable, BakersLtable },
	{ BakersTable, BakersLtable },
	CardMoveNoHelp,
	0,
	BakersCheckTable,
	0,
    },
};

#define NMOVES	(sizeof (BakersMoves) / sizeof (BakersMoves[0]))

static CardBool
BakersNormalize (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    CardIndex	    card;
    CardType	    *cp;

    for (stack = BakersPile; stack <= BakersLpile; stack++)
    {
	sp = CardStackPtr (stack);
	card = sp->last;
	if (card == NullIndex)
	    return False;
	cp = CardPtr (card);
	if (Rank (cp->card) != KING)
	    return False;
    }
    return True;
}

RulesType   bakersDozenGame = {
    MakeCard (0, CLUB, 1), MakeCard (0, SPADE, KING), 1, BakersNstack,
    BakersNormalize,
    BakersInitialize,
    CardIsInOrder,
    BakersSetup,
    NMOVES,
    BakersMoves
};
