/*
 * $Id: forty.c,v 1.6 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"

#define FortyThievesDeck	0
#define FortyThievesDiscard	1
#define FortyThievesPile	(FortyThievesDiscard + 1)
#define FortyThievesNpile	8
#define FortyThievesLpile	(FortyThievesPile + FortyThievesNpile - 1)
#define FortyThievesTable	(FortyThievesLpile + 1)
#define FortyThievesNtable	10
#define FortyThievesLtable	(FortyThievesTable + FortyThievesNtable - 1)
#define FortyThievesNstack	(FortyThievesLtable + 1)

CardSetupType	FortyThievesSetup[] = {
    { { 9, 0 }, { 0, 0 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartDown },
    { { 8, 0 }, { 1, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartUp },

    { { 0, 0 }, { 2, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 1, 0 }, { 3, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 2, 0 }, { 4, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 3, 0 }, { 5, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 4, 0 }, { 6, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 5, 0 }, { 7, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 6, 0 }, { 8, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 7, 0 }, { 9, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },

    { { 0, 1 }, { 0, 1 }, CardNone, CardDisplayAll, False, False, 4, CardStartUp },
    { { 1, 1 }, { 1, 1 }, CardNone, CardDisplayAll, False, False, 4, CardStartUp },
    { { 2, 1 }, { 2, 1 }, CardNone, CardDisplayAll, False, False, 4, CardStartUp },
    { { 3, 1 }, { 3, 1 }, CardNone, CardDisplayAll, False, False, 4, CardStartUp },
    { { 4, 1 }, { 4, 1 }, CardNone, CardDisplayAll, False, False, 4, CardStartUp },
    { { 5, 1 }, { 5, 1 }, CardNone, CardDisplayAll, False, False, 4, CardStartUp },
    { { 6, 1 }, { 6, 1 }, CardNone, CardDisplayAll, False, False, 4, CardStartUp },
    { { 7, 1 }, { 7, 1 }, CardNone, CardDisplayAll, False, False, 4, CardStartUp },
    { { 8, 1 }, { 8, 1 }, CardNone, CardDisplayAll, False, False, 4, CardStartUp },
    { { 9, 1 }, { 9, 1 }, CardNone, CardDisplayAll, False, False, 4, CardStartUp },
};

static CardBool FortyThievesSelectTableUnordered (CardStackIndex stack,
        CardIndex *first, CardIndex *last) GAME_SECT;
static CardProgressType FortyThievesCheckTable (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType FortyThievesCheckEmpty (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType FortyThievesCheckPile (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static void FortyThievesMoveDeal (CardStackIndex src,
        CardStackIndex dst, CardIndex last) GAME_SECT;
static CardBool FortyThievesNormalize (void) GAME_SECT;

// Only allow moving one card, and only if it's not already on
// an in-suit-order card
static CardBool
FortyThievesSelectTableUnordered (CardStackIndex   stack,
		   CardIndex	    *first,
		   CardIndex	    *last)
{
    CardStackType   *sp = CardStackPtr (stack);
    CardIndex	    card;
    CardType	    *cp;

    card = sp->last;
    *first = card;
    *last = card;
    cp = CardPtr (card);
    if (cp->prev != NullIndex && CardIsInSuitOrder (card, cp->prev))
        return False;
    return True;
}

static CardProgressType
FortyThievesCheckTable (CardStackIndex	src, 
		 CardStackIndex dst,
		 CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (CardIsInSuitOrder (card, dp->last))
	return CardProgressSome;
    return CardProgressCant;
}

static CardProgressType
FortyThievesCheckEmpty (CardStackIndex	src,
		      CardStackIndex	dst,
		      CardIndex		card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
	return CardProgressSome;
    return CardProgressCant;
}

static CardProgressType
FortyThievesCheckPile (CardStackIndex	src, 
		CardStackIndex	dst,
		CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
    {
	if (Rank (CardPtr (card)->card) == 1)
	    return CardProgressSome;
    }
    else
    {
	if (CardIsInSuitOrder (dp->last, card))
	    return CardProgressSome;
    }
    return CardProgressCant;
}

static void
FortyThievesMoveDeal (CardStackIndex	src,
	      CardStackIndex	dst,
	      CardIndex		last)
{
    CardIndex	card;

    card = CardStackPtr (src)->last;
    CardMove (src, card, dst, True);
    CardTurn (dst, card, True, True);
}

CardMoveType	FortyThievesMoves[] = {
    /* 0    Move card from table to pile */
    {	"table to pile",
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesPile, FortyThievesLpile },
	CardMoveAuto,
	0,
	FortyThievesCheckPile,
	0,
    },
    /* 1    Move card from discard to pile */
    {	"discard to pile",
	{ FortyThievesDiscard, FortyThievesDiscard },
	{ FortyThievesDiscard, FortyThievesDiscard },
	{ FortyThievesPile, FortyThievesLpile },
	CardMoveAuto,
	0,
	FortyThievesCheckPile,
	0,
    },
    /* 2    Move card from discard to table */
    {	"discard to table",
	{ FortyThievesDiscard, FortyThievesDiscard },
	{ FortyThievesDiscard, FortyThievesDiscard },
	{ FortyThievesTable, FortyThievesLtable },
	CardMoveHelpful,
	0,
	FortyThievesCheckTable,
	0,
    },
    /* 3    Move card from table to table */
    {	"table to table (if not ordered)",
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	CardMoveHelpful,
	FortyThievesSelectTableUnordered,
	FortyThievesCheckTable,
	0,
    },
    /* 4    Move card from table to empty table */
    {	"table to empty table (if not ordered)",
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	CardMoveHelpful,
	FortyThievesSelectTableUnordered,
	FortyThievesCheckEmpty,
	0,
    },
    /* 5    Move card from discard to empty table */
    {	"discard to empty table",
	{ FortyThievesDiscard, FortyThievesDiscard },
	{ FortyThievesDiscard, FortyThievesDiscard },
	{ FortyThievesTable, FortyThievesLtable },
	CardMoveHelpful,
	0,
	FortyThievesCheckEmpty,
	0,
    },
    /* 6    Move card from pile to table */
    {	"pile to table",
	{ FortyThievesPile, FortyThievesLpile },
	{ FortyThievesPile, FortyThievesLpile },
	{ FortyThievesTable, FortyThievesLtable },
	CardMoveNoHelp,
	0,
	FortyThievesCheckTable,
	0,
    },
    /* 7    Move card from pile to empty table */
    {	"pile to empty table",
	{ FortyThievesPile, FortyThievesLpile },
	{ FortyThievesPile, FortyThievesLpile },
	{ FortyThievesTable, FortyThievesLtable },
	CardMoveNoHelp,
	0,
	FortyThievesCheckEmpty,
	0,
    },
    /* 8    Move card from table to table */
    {	"table to table",
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	CardMoveNoHelp,
	0,
	FortyThievesCheckTable,
	0,
    },
    /* 9    Move card from table to empty table */
    {	"table to empty table",
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	CardMoveNoHelp,
	0,
	FortyThievesCheckEmpty,
	0,
    },
    /* 10    Move card from table to empty discard */
    {	"table to empty discard",
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesDiscard, FortyThievesDiscard },
	CardMoveNoHelp,
	0,
	FortyThievesCheckEmpty,
	0,
    },
    /* 11    Move card from table to empty deck */
    {	"table to empty deck",
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesTable, FortyThievesLtable },
	{ FortyThievesDeck, FortyThievesDeck },
	CardMoveNoHelp,
	0,
	FortyThievesCheckEmpty,
	0,
    },
    /* 12   Deal */
    {	"deal",
	{ FortyThievesDeck, FortyThievesDeck },
	{ FortyThievesDeck, FortyThievesDeck },
	{ FortyThievesDiscard, FortyThievesDiscard },
	CardMoveHelpful,
	0,
	0,
	FortyThievesMoveDeal,
    },
};

#define NMOVES	(sizeof (FortyThievesMoves) / sizeof (FortyThievesMoves[0]))

static CardBool
FortyThievesNormalize (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    CardIndex	    card;
    CardType	    *cp;

    for (stack = FortyThievesPile; stack <= FortyThievesLpile; stack++)
    {
	sp = CardStackPtr (stack);
	card = sp->last;
	if (card == NullIndex)
	    return False;
	cp = CardPtr (card);
	if (Rank (cp->card) != KING)
	    return False;
    }
    return True;
}

RulesType   fortyThievesGame = {
    MakeCard (0, CLUB, 1), MakeCard (0, SPADE, KING), 2, FortyThievesNstack,
    FortyThievesNormalize,
    0,
    0,
    FortyThievesSetup,
    NMOVES,
    FortyThievesMoves
};
