/*
 * $Id: freecell.c,v 1.5 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"

#define FreecellDeck	0
#define FreecellTemp	1
#define FreecellNtemp	4
#define FreecellLtemp	(FreecellTemp + FreecellNtemp - 1)
#define FreecellPile	(FreecellLtemp + 1)
#define FreecellNpile	4
#define FreecellLpile	(FreecellPile + FreecellNpile - 1)
#define FreecellTable	(FreecellLpile + 1)
#define FreecellNtable	8
#define FreecellLtable	(FreecellTable + FreecellNtable - 1)
#define FreecellNstack	(FreecellLtable + 1)

CardSetupType	FreecellSetup[] =  {
    { { 0, 0 }, { 0, 0 }, CardEmpty, CardDisplayNone, False, False, 0, CardStartDown },

    { { 0, 0 }, { 0, 0 }, CardBlank, CardDisplayTop, False, False, 0, CardStartUp },
    { { 1, 0 }, { 1, 0 }, CardBlank, CardDisplayTop, False, False, 0, CardStartUp },
    { { 2, 0 }, { 2, 0 }, CardBlank, CardDisplayTop, False, False, 0, CardStartUp },
    { { 3, 0 }, { 3, 0 }, CardBlank, CardDisplayTop, False, False, 0, CardStartUp },

    { { 4, 0 }, { 4, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 5, 0 }, { 5, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 6, 0 }, { 6, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 7, 0 }, { 7, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },

    { { 0, 1 }, { 0, 1 }, CardNone, CardDisplayPack, False, True, 7, CardStartUp },
    { { 1, 1 }, { 1, 1 }, CardNone, CardDisplayPack, False, True, 7, CardStartUp },
    { { 2, 1 }, { 2, 1 }, CardNone, CardDisplayPack, False, True, 7, CardStartUp },
    { { 3, 1 }, { 3, 1 }, CardNone, CardDisplayPack, False, True, 7, CardStartUp },
    { { 4, 1 }, { 4, 1 }, CardNone, CardDisplayPack, False, True, 6, CardStartUp },
    { { 5, 1 }, { 5, 1 }, CardNone, CardDisplayPack, False, True, 6, CardStartUp },
    { { 6, 1 }, { 6, 1 }, CardNone, CardDisplayPack, False, True, 6, CardStartUp },
    { { 7, 1 }, { 7, 1 }, CardNone, CardDisplayPack, False, True, 6, CardStartUp },
};

void FreecellInitialize (void) GAME_SECT;
static int FreecellCountTemp (void) GAME_SECT;
static CardBool FreecellSelectTable (CardStackIndex, CardIndex*, CardIndex*) GAME_SECT;
static CardProgressType FreecellCheckTable (CardStackIndex, CardStackIndex, CardIndex) GAME_SECT;
static CardProgressType FreecellCheckTablePartial (CardStackIndex, CardStackIndex, CardIndex) GAME_SECT;
static CardProgressType FreecellCheckEmptyTable (CardStackIndex, CardStackIndex, CardIndex) GAME_SECT;
static CardProgressType FreecellCheckEmptyTablePartial (CardStackIndex, CardStackIndex, CardIndex) GAME_SECT;
static CardProgressType FreecellCheckPile (CardStackIndex, CardStackIndex, CardIndex) GAME_SECT;
static CardProgressType FreecellCheckTemp (CardStackIndex, CardStackIndex, CardIndex) GAME_SECT;
static CardBool FreecellNormalize (void) GAME_SECT;


// special random number generator for freecell game numbers
static unsigned long next = 1;

static int rand(void)
{
    next = (next * 214013 + 2531011L);
    return (next >> 16) & 0x7fff;
}

static void srand(unsigned int seed)
{
    next = seed;
}


void
FreecellInitialize (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    CardType	    *cp;

    // Freecell shuffle details
#define K_SUIT(card)      ((card) % 4)
#define K_VALUE(card)     ((card) / 4)
#define CONV_CARD(card)   (K_SUIT(card) * 13 + K_VALUE(card))

#define MAXPOS  8
#define MAXCOL  8
    int cards[MAXCOL][MAXPOS];
    // srand(1) -> col[0] = {41, 49, 7, 12, 11, 21, 23} = Jd Kd 2s 4c 3s 6d 6s
    // patience equiv = 27, 29, 50, 4, 51, 22, 54

    int  i, j, col, pos;
    int  wLeft = 52;            // cards left to be chosen in shuffle
    int  deck[52];              // deck of 52 unique cards

    // Implement freecell game number shuffle
    // Note: algorithm freely available
    //     http://home.earthlink.net/~fomalhaut/mshuffle.txt

    srand(game.gameSeed);

    for (col = 0; col < MAXCOL; col++)
        for (pos = 0; pos < MAXPOS; pos++)
            cards[col][pos] = -1;

    for (i = 0; i < 52; i++)      // put unique card in each deck loc
        deck[i] = i;

    for (i = 0; i < 52; i++)
    {
        j = rand() % wLeft;
        cards[(i%8)][i/8] = deck[j];
        deck[j] = deck[--wLeft];
    }

    // convert to patience standards
    for (stack = FreecellTable, col=0; stack <= FreecellLtable; stack++, col++)
    {
        sp = CardStackPtr (stack);

        // based on patience shuffle algorithm
        sp->first = CONV_CARD(cards[col][0]);
        for (i = 0; cards[col][i] != -1; i++)
        {
            cp = CardPtr (CONV_CARD(cards[col][i]));
            cp->prev = (i==0)? NullIndex: CONV_CARD(cards[col][i-1]);
            if (cards[col][i+1] != -1)
                cp->next = CONV_CARD(cards[col][i+1]);
            else
                cp->next = NullIndex;
        }
        sp->last = CONV_CARD(cards[col][i-1]);
        sp->dirty = True;
    }
}

// TODO All of the "cell" games should include a count of empty table
// columns as well, at least when the rules allow arbitrary cards in an
// empty table slot.  The number of empty cells is multiplied by the number
// of empty table columns (called a "supermove" in freecell)
static int
FreecellCountTemp (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    int		    nempty = 0;

    for (stack = FreecellTemp; stack <= FreecellLtemp; stack++)
    {
	sp = CardStackPtr (stack);
	if (sp->last == NullIndex)
	    nempty++;
    }
    return nempty;
}

static CardBool
FreecellSelectTable (CardStackIndex stack,
		     CardIndex	    *first,
		     CardIndex	    *last)
{
    CardStackType   *sp = CardStackPtr (stack);
    CardIndex	    card;
    CardType	    *cp;
    int		    nempty = FreecellCountTemp ();
    int		    nstack = 0;

    card = sp->last;
    *first = card;
    while (card != NullIndex)
    {
	nstack++;
	if (nstack > nempty)
	    break;
	cp = CardPtr (card);
	if (cp->prev == NullIndex)
	    break;
	if (!CardIsInAlternatingSuitOrder (card, cp->prev))
	    break;
	card = cp->prev;
    }
    *last = card;
    return True;
}

static CardProgressType
FreecellCheckTable (CardStackIndex    src, 
		  CardStackIndex    dst,
		  CardIndex	    card)
{
    CardType	    *cp = CardPtr (card);
    CardStackType   *dp = CardStackPtr (dst);

    if (CardIsInAlternatingSuitOrder (card, dp->last))
    {
        // don't hint at moving part of a build somewhere else
        // (but still allow it)
	if (CardIsInAlternatingSuitOrder (card, cp->prev))
            return CardProgressNone;
        return CardProgressSome;
    }
    return CardProgressCant;
}

// same as above, but provides help for partial builds if no other help is
// appropriate
static CardProgressType
FreecellCheckTablePartial (CardStackIndex    src, 
		  CardStackIndex    dst,
		  CardIndex	    card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (CardIsInAlternatingSuitOrder (card, dp->last))
        return CardProgressSome;
    return CardProgressCant;
}

static CardProgressType
FreecellCheckEmptyTable (CardStackIndex	src,
		       CardStackIndex	dst,
		       CardIndex	card)
{
    CardType	    *cp = CardPtr (card);
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
    {
        // don't hint at moving from empty to empty, or part of a build
        // somewhere else (but still allow it)
        if ( (FreecellTable <= src && src <= FreecellLtable &&
                    cp->prev == NullIndex) ||
                CardIsInAlternatingSuitOrder (card, cp->prev))
            return CardProgressNone;
        return CardProgressSome;
    }
    return CardProgressCant;
}

// same as above, but provides help for partial builds if no other help is
// appropriate
static CardProgressType
FreecellCheckEmptyTablePartial (CardStackIndex	src,
		       CardStackIndex	dst,
		       CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
        return CardProgressSome;
    return CardProgressCant;
}

static CardProgressType
FreecellCheckPile (CardStackIndex src, 
		 CardStackIndex dst,
		 CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
    {
	if (Rank (CardPtr (card)->card) == 1)
	    return CardProgressSome;
    }
    else
    {
	if (CardIsInSuitOrder (dp->last, card))
	    return CardProgressSome;
    }
    return CardProgressCant;
}

static CardProgressType
FreecellCheckTemp (CardStackIndex src, 
		 CardStackIndex dst,
		 CardIndex	card)
{
    CardType	    *cp = CardPtr (card);
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
    {
        // don't hint at moving part of a build somewhere else
        // (but still allow it)
	if (CardIsInAlternatingSuitOrder (card, cp->prev))
            return CardProgressNone;
        return CardProgressSome;
    }
    return CardProgressCant;
}

CardMoveType FreecellMoves[] = {
    /* 0    Move card from table to pile */
    {	"table to pile",
	{ FreecellTable, FreecellLtable },
	{ FreecellTable, FreecellLtable },
	{ FreecellPile, FreecellLpile },
	CardMoveAuto,
	0,
	FreecellCheckPile,
	0,
    },
    /* 1    Move card from temp to pile */
    {	"temp to pile",
	{ FreecellTemp, FreecellLtemp },
	{ FreecellTemp, FreecellLtemp },
	{ FreecellPile, FreecellLpile },
	CardMoveAuto,
	0,
	FreecellCheckPile,
	0,
    },
    /* 2    Move card from table to table */
    {	"table to table",
	{ FreecellTable, FreecellLtable },
	{ FreecellTable, FreecellLtable },
	{ FreecellTable, FreecellLtable },
	CardMoveHelpful,
	FreecellSelectTable,
	FreecellCheckTable,
	0,
    },
    /* 3    Move card from table to empty table */
    {	"table to empty table",
	{ FreecellTable, FreecellLtable },
	{ FreecellTable, FreecellLtable },
	{ FreecellTable, FreecellLtable },
	CardMoveHelpful,
	FreecellSelectTable,
	FreecellCheckEmptyTable,
	0,
    },
    /* 4    Move card from temp to table */
    {	"temp to table",
	{ FreecellTemp, FreecellLtemp },
	{ FreecellTemp, FreecellLtemp },
	{ FreecellTable, FreecellLtable },
	CardMoveHelpful,
	0,
	FreecellCheckTable,
	0,
    },
    /* 5    Move card from temp to empty table */
    {	"temp to empty table",
	{ FreecellTemp, FreecellLtemp },
	{ FreecellTemp, FreecellLtemp },
	{ FreecellTable, FreecellLtable },
	CardMoveHelpful,
	0,
	FreecellCheckEmptyTable,
	0,
    },
    /* 6    Move card from table to temp */
    {	"table to temp",
	{ FreecellTable, FreecellLtable },
	{ FreecellTable, FreecellLtable },
	{ FreecellTemp, FreecellLtemp },
	CardMoveHelpful,
	0,
	FreecellCheckTemp,
	0,
    },
    /* 7    Move card from pile to table */
    {	"pile to table",
	{ FreecellPile, FreecellLpile },
	{ FreecellPile, FreecellLpile },
	{ FreecellTable, FreecellLtable },
	CardMoveNoHelp,
	0,
	FreecellCheckTable,
	0,
    },
    /* 8    Move card from pile to empty table */
    {	"pile to empty table",
	{ FreecellPile, FreecellLpile },
	{ FreecellPile, FreecellLpile },
	{ FreecellTable, FreecellLtable },
	CardMoveNoHelp,
	0,
	FreecellCheckEmptyTable,
	0,
    },
    /* 9    Move partial build from table to table */
    {	"partial table to table",
	{ FreecellTable, FreecellLtable },
	{ FreecellTable, FreecellLtable },
	{ FreecellTable, FreecellLtable },
	CardMoveHelpful,
	FreecellSelectTable,
	FreecellCheckTablePartial,
	0,
    },
    /* 3    Move partial build from table to empty table */
    {	"partial table to empty table",
	{ FreecellTable, FreecellLtable },
	{ FreecellTable, FreecellLtable },
	{ FreecellTable, FreecellLtable },
	CardMoveHelpful,
	FreecellSelectTable,
	FreecellCheckEmptyTablePartial,
	0,
    },
};

#define NMOVES	(sizeof (FreecellMoves) / sizeof (FreecellMoves[0]))

static CardBool
FreecellNormalize (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    CardIndex	    card;
    CardType	    *cp;

    for (stack = FreecellPile; stack <= FreecellLpile; stack++)
    {
	sp = CardStackPtr (stack);
	card = sp->last;
	if (card == NullIndex)
	    return False;
	cp = CardPtr (card);
	if (Rank (cp->card) != KING)
	    return False;
    }
    return True;
}

RulesType   freecellGame = {
    MakeCard (0, CLUB, 1), MakeCard (0, SPADE, KING), 1, FreecellNstack,
    FreecellNormalize,
    FreecellInitialize,
    CardIsInAlternatingSuitOrder,
    FreecellSetup,
    NMOVES,
    FreecellMoves
};
