/*
 * $Id: gypsy.c,v 1.6 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"

#define GypsyDeck	0
#define GypsyPile	(GypsyDeck + 1)
#define GypsyNpile	8
#define GypsyLpile	(GypsyPile + GypsyNpile - 1)
#define GypsyTable	(GypsyLpile + 1)
#define GypsyNtable	8
#define GypsyLtable	(GypsyTable + GypsyNtable - 1)
#define GypsyNstack	(GypsyLtable + 1)

CardSetupType	GypsySetup[] =  {
    { { 0, 0 }, { 0, 0 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartDown },

    { { 1, 0 }, { 1, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 2, 0 }, { 2, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 3, 0 }, { 3, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 4, 0 }, { 4, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 5, 0 }, { 5, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 6, 0 }, { 6, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 7, 0 }, { 7, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 8, 0 }, { 8, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },

    { { 1, 1 }, { 1, 1 }, CardNone, CardDisplayPack, False, False, 3, CardStartLastUp },
    { { 2, 1 }, { 2, 1 }, CardNone, CardDisplayPack, False, False, 3, CardStartLastUp },
    { { 3, 1 }, { 3, 1 }, CardNone, CardDisplayPack, False, False, 3, CardStartLastUp },
    { { 4, 1 }, { 4, 1 }, CardNone, CardDisplayPack, False, False, 3, CardStartLastUp },
    { { 5, 1 }, { 5, 1 }, CardNone, CardDisplayPack, False, False, 3, CardStartLastUp },
    { { 6, 1 }, { 6, 1 }, CardNone, CardDisplayPack, False, False, 3, CardStartLastUp },
    { { 7, 1 }, { 7, 1 }, CardNone, CardDisplayPack, False, False, 3, CardStartLastUp },
    { { 8, 1 }, { 8, 1 }, CardNone, CardDisplayPack, False, False, 3, CardStartLastUp },


};
static CardBool GypsySelectTable (CardStackIndex stack,
        CardIndex *first, CardIndex *last) GAME_SECT;
static CardProgressType GypsyCheckTable (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType GypsyCheckEmpty (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType GypsyCheckPile (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static void GypsyMoveDeal (CardStackIndex src, CardStackIndex dst,
        CardIndex last) GAME_SECT;
static CardBool GypsyNormalize (void) GAME_SECT;

static CardBool
GypsySelectTable (CardStackIndex stack,
		   CardIndex	    *first,
		   CardIndex	    *last)
{
    CardStackType   *sp = CardStackPtr (stack);

    *first = sp->last;
    *last = CardInReverseAlternatingSuitOrder (sp->last);
    return True;
}

static CardProgressType
GypsyCheckTable (CardStackIndex  src,
		  CardStackIndex    dst,
		  CardIndex	    card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (CardIsInAlternatingSuitOrder (card, dp->last))
	return CardProgressSome;
    return CardProgressCant;
}

static CardProgressType
GypsyCheckEmpty (CardStackIndex  src,
		  CardStackIndex    dst,
		  CardIndex	    card)
{
    CardStackType   *dp = CardStackPtr (dst);
    CardStackType   *sp = CardStackPtr (src);

    if (dp->last != NullIndex)
	return CardProgressCant;
    // Don't suggest useless single-stack to empty moves
    if (sp->first == card)
        return CardProgressNoHint;
    return CardProgressSome;
}

static CardProgressType
GypsyCheckPile (CardStackIndex src,
		 CardStackIndex dst,
		 CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
    {
	if (Rank (CardPtr (card)->card) == 1)
	  return CardProgressSome;
    }
    else
    {
	if (CardIsInSuitOrder (dp->last, card))
	    return CardProgressSome;
    }
    return CardProgressCant;
}

static void
GypsyMoveDeal (CardStackIndex  src,
		CardStackIndex  dst,
		CardIndex	last)
{
    CardStackIndex  t;

    for (t = GypsyTable; t <= GypsyLtable; t++)
    {
	if (CardStackPtr (GypsyDeck)->last == NullIndex)
	    return;
	CardMove (GypsyDeck, CardStackPtr (GypsyDeck)->last,
		  t, True);
	CardTurn (t, CardStackPtr (t)->last, True, True);
    }
}

CardMoveType GypsyMoves[] = {
    /* 0    Move card from table to pile */
    {	"table to pile",
	{ GypsyTable, GypsyLtable },
	{ GypsyTable, GypsyLtable },
	{ GypsyPile, GypsyLpile },
	CardMoveAuto,
	0,
	GypsyCheckPile,
	0,
    },
    /* 1    Move card from table to table */
    {	"table to table",
	{ GypsyTable, GypsyLtable },
	{ GypsyTable, GypsyLtable },
	{ GypsyTable, GypsyLtable },
	CardMoveHelpful,
	GypsySelectTable,
	GypsyCheckTable,
	0,
    },
    /* 2    Move card from table to empty */
    {	"table to empty",
	{ GypsyTable, GypsyLtable },
	{ GypsyTable, GypsyLtable },
	{ GypsyTable, GypsyLtable },
	CardMoveHelpful,
	GypsySelectTable,
	GypsyCheckEmpty,
	0,
    },

    /* 3    Move card from pile to table */ 
    {	"pile to table",
	{ GypsyPile, GypsyPile },
	{ GypsyPile, GypsyPile },
	{ GypsyTable, GypsyLtable },
	CardMoveNoHelp,
	0,
	GypsyCheckTable,
	0,
    },
    /* 4    deal */
    {	"deal",
	{ GypsyDeck, GypsyDeck },
	{ GypsyDeck, GypsyDeck },
	{ GypsyTable, GypsyLtable },
	CardMoveHelpful,
	0,
	0,
	GypsyMoveDeal,
    },
};

#define NMOVES	(sizeof (GypsyMoves) / sizeof (GypsyMoves[0]))

static CardBool
GypsyNormalize (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    CardIndex	    card;
    CardType	    *cp;

    // Turn up any face down cards revealed by playing other cards
    for (stack = GypsyTable; stack <= GypsyLtable; stack++)
    {
	sp = CardStackPtr (stack);
	card = sp->last;
	if (card != NullIndex)
	{
	    cp = CardPtr (card);
	    if (!cp->face)
		CardTurn (stack, card, True, True);
	}
    }

    for (stack = GypsyPile; stack <= GypsyLpile; stack++)
    {
	card = CardStackPtr (stack)->last;
	if (card == NullIndex)
	    return False;
	if (Rank (CardPtr (card)->card) != KING)
	    return False;
    }
    return True;
}

RulesType   gypsyGame = {
    MakeCard (0, CLUB, 1), MakeCard (0, SPADE, KING), 2, GypsyNstack,
    GypsyNormalize,
    0,
    CardIsInAlternatingSuitOrder,
    GypsySetup,
    NMOVES,
    GypsyMoves
};
