/*
 * $Id: milligan.c,v 1.5 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"

#define MilliganDeck	0
#define MilliganPile	(MilliganDeck + 1)
#define MilliganNpile	8
#define MilliganLpile	(MilliganPile + MilliganNpile - 1)
#define MilliganTable	(MilliganLpile + 1)
#define MilliganNtable	8
#define MilliganLtable	(MilliganTable + MilliganNtable - 1)
#define MilliganTemp	(MilliganLtable + 1)
#define MilliganNstack	(MilliganTemp + 1)

CardSetupType	MilliganSetup[] =  {
    { { 0, 0 }, { 0, 0 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartDown },

    { { 1, 0 }, { 1, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 2, 0 }, { 2, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 3, 0 }, { 3, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 4, 0 }, { 4, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 5, 0 }, { 5, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 6, 0 }, { 6, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 7, 0 }, { 7, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 8, 0 }, { 8, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },

    { { 1, 1 }, { 1, 1 }, CardNone, CardDisplayPack, False, False, 1, CardStartLastUp },
    { { 2, 1 }, { 2, 1 }, CardNone, CardDisplayPack, False, False, 1, CardStartLastUp },
    { { 3, 1 }, { 3, 1 }, CardNone, CardDisplayPack, False, False, 1, CardStartLastUp },
    { { 4, 1 }, { 4, 1 }, CardNone, CardDisplayPack, False, False, 1, CardStartLastUp },
    { { 5, 1 }, { 5, 1 }, CardNone, CardDisplayPack, False, False, 1, CardStartLastUp },
    { { 6, 1 }, { 6, 1 }, CardNone, CardDisplayPack, False, False, 1, CardStartLastUp },
    { { 7, 1 }, { 7, 1 }, CardNone, CardDisplayPack, False, False, 1, CardStartLastUp },
    { { 8, 1 }, { 8, 1 }, CardNone, CardDisplayPack, False, False, 1, CardStartLastUp },

    { { 0, 1 }, { 0, 1 }, CardEmpty, CardDisplayPack, False, False, 0, CardStartLastUp },
};

static CardBool MilliganSelectTable (CardStackIndex stack,
        CardIndex *first, CardIndex *last) GAME_SECT;
static CardProgressType MilliganCheckTable (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType MilliganCheckEmpty (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType MilliganCheckPile (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType MilliganCheckTemp (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static void MilliganMoveDeal (CardStackIndex src, CardStackIndex dst,
        CardIndex last) GAME_SECT;
static CardBool MilliganNormalize (void) GAME_SECT;

static CardBool
MilliganSelectTable (CardStackIndex stack,
		   CardIndex	    *first,
		   CardIndex	    *last)
{
    CardStackType   *sp = CardStackPtr (stack);

    *first = sp->last;
    *last = CardInReverseAlternatingSuitOrder (sp->last);
    return True;
}

static CardProgressType
MilliganCheckTable (CardStackIndex  src,
		  CardStackIndex    dst,
		  CardIndex	    card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (CardIsInAlternatingSuitOrder (card, dp->last))
	return CardProgressSome;
    return CardProgressCant;
}

static CardProgressType
MilliganCheckEmpty (CardStackIndex  src,
		  CardStackIndex    dst,
		  CardIndex	    card)
{
    CardStackType   *dp = CardStackPtr (dst);
    CardStackType   *sp = CardStackPtr (src);
    CardType	    *cp = CardPtr (card);

    if (dp->last != NullIndex || Rank (cp->card) != KING)
        return CardProgressCant;
    // Don't suggest useless King-stack to empty moves
    if (src != MilliganTemp && sp->first == card)
        return CardProgressNoHint;
    return CardProgressSome;
}

static CardProgressType
MilliganCheckPile (CardStackIndex src,
		 CardStackIndex dst,
		 CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
    {
	if (Rank (CardPtr (card)->card) == 1)
	    return CardProgressSome;
    }
    else
    {
	if (CardIsInSuitOrder (dp->last, card))
	    return CardProgressSome;
    }
    return CardProgressCant;
}

static CardProgressType
MilliganCheckTemp (CardStackIndex   src,
		  CardStackIndex    dst,
		  CardIndex	    card)
{
    CardStackType   *deckp, *dp;

    deckp = CardStackPtr (MilliganDeck);
    dp = CardStackPtr (dst);
    if (deckp->last == NullIndex && dp->last == NullIndex)
        return CardProgressSome;
    return CardProgressCant;
}

static void
MilliganMoveDeal (CardStackIndex  src,
		CardStackIndex  dst,
		CardIndex	last)
{
    CardStackIndex  t;

    for (t = MilliganTable; t <= MilliganLtable; t++)
    {
	if (CardStackPtr (MilliganDeck)->last == NullIndex)
	    return;
	CardMove (MilliganDeck, CardStackPtr (MilliganDeck)->last,
		  t, True);
	CardTurn (t, CardStackPtr (t)->last, True, True);
    }
}

CardMoveType MilliganMoves[] = {
    /* 0    Move card from temp to pile */
    {	"temp to pile",
	{ MilliganTemp, MilliganTemp },
	{ MilliganTemp, MilliganTemp },
	{ MilliganPile, MilliganLpile },
	CardMoveAuto,
	0,
	MilliganCheckPile,
	0,
    },
    /* 1    Move card from table to pile */
    {	"table to pile",
	{ MilliganTable, MilliganLtable },
	{ MilliganTable, MilliganLtable },
	{ MilliganPile, MilliganLpile },
	CardMoveAuto,
	0,
	MilliganCheckPile,
	0,
    },
    /* 2    Move card from temp to table */
    {	"temp to table",
	{ MilliganTemp, MilliganTemp },
	{ MilliganTemp, MilliganTemp },
	{ MilliganTable, MilliganLtable },
	CardMoveHelpful,
	MilliganSelectTable,
	MilliganCheckTable,
	0,
    },
    /* 3    Move card from table to table */
    {	"table to table",
	{ MilliganTable, MilliganLtable },
	{ MilliganTable, MilliganLtable },
	{ MilliganTable, MilliganLtable },
	CardMoveHelpful,
	MilliganSelectTable,
	MilliganCheckTable,
	0,
    },
    /* 4    Move card from temp to empty */
    {	"temp to empty",
	{ MilliganTemp, MilliganTemp },
	{ MilliganTemp, MilliganTemp },
	{ MilliganTable, MilliganLtable },
	CardMoveHelpful,
	MilliganSelectTable,
	MilliganCheckEmpty,
	0,
    },
    /* 5    Move card from table to empty */
    {	"table to empty",
	{ MilliganTable, MilliganLtable },
	{ MilliganTable, MilliganLtable },
	{ MilliganTable, MilliganLtable },
	CardMoveHelpful,
	MilliganSelectTable,
	MilliganCheckEmpty,
	0,
    },
    /* 6    Move card from table to (empty) temp */
    {	"table to temp",
	{ MilliganTable, MilliganLtable },
	{ MilliganTable, MilliganLtable },
	{ MilliganTemp, MilliganTemp },
	CardMoveHelpful,
	MilliganSelectTable,
	MilliganCheckTemp,
	0,
    },
    /* 7    deal */
    {	"deal",
	{ MilliganDeck, MilliganDeck },
	{ MilliganDeck, MilliganDeck },
	{ MilliganTable, MilliganLtable },
	CardMoveHelpful,
	0,
	0,
	MilliganMoveDeal,
    },
};

#define NMOVES	(sizeof (MilliganMoves) / sizeof (MilliganMoves[0]))

static CardBool
MilliganNormalize (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    CardIndex	    card;
    CardType	    *cp;

    for (stack = MilliganPile; stack <= MilliganLpile; stack++)
    {
	sp = CardStackPtr (stack);
	card = sp->last;
	if (card == NullIndex)
	    return False;
	cp = CardPtr (card);
	if (Rank (cp->card) != KING)
	    return False;
    }
    return True;
}

RulesType   milliganGame = {
    MakeCard (0, CLUB, 1), MakeCard (0, SPADE, KING), 2, MilliganNstack,
    MilliganNormalize,
    0,
    CardIsInAlternatingSuitOrder,
    MilliganSetup,
    NMOVES,
    MilliganMoves
};
