/*
 * $Id: patience.h,v 1.11 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"

extern PatS    cardWidth;
extern PatS    cardHeight;
extern PatS    cardSpaceX;
extern PatS    cardSpaceY;
extern PatS    cardLabel;
extern PatS    cardRadius;
extern PatS    cardOriginX;
extern PatS    cardOriginY;
extern PatS    cardBackOffset;
extern PatS    cardPackOffset;
extern PatS    cardLabelWidth;
extern PatS    cardLabelHeight;
extern PatS    xOffset;
extern PatS    yOffset;
extern PatS    maxCardWidth;
extern PatS    maxCardHeight;
extern UInt16  num_games;
extern UInt32  romVersion, sysVersion30, sysVersion33, sysVersion35,
       sysVersion40, sysVersion50;
extern UInt32  screen_depth;
extern UInt32  form_width;
extern UInt32  form_height;
extern ColumnInfo column[MAX_ROWCOL];

// High-res mode(s)
typedef enum {
    noHR=0,
    PalmOS5HR,
    SonyHR
} HRMode;
// Sony High-res mode
extern UInt16 hrRefNum;
extern HRMode hrMode;
// HandEra QVGA mode
extern Boolean qvgaMode;
// Sony silkscreen manager
extern Boolean silkMode;
extern Boolean window_resized;
// PalmOS dynamic input area
extern Boolean diaMode;

#define GetObjectPtrById(frm,objid) \
        FrmGetObjectPtr((frm), FrmGetObjectIndex((frm), (objid)))
#define GetObjectBounds(frm,objid,r) \
        FrmGetObjectBounds((frm), FrmGetObjectIndex((frm), (objid)), (r))
#define GetLabelById(frm,objid) \
        CtlGetLabel(FrmGetObjectPtr((frm), FrmGetObjectIndex((frm), (objid))))

// Color definitions
// see also http://www.tayloredmktg.com/rgb
#define COLOR_black	{ 0, 0x00, 0x00, 0x00 }
#define COLOR_grey	{ 0, 0x66, 0x66, 0x66 }
#define COLOR_lightGrey	{ 0, 0xaa, 0xaa, 0xaa }
#define COLOR_silver	{ 0, 0xc0, 0xc0, 0xc0 }
#define COLOR_white	{ 0, 0xff, 0xff, 0xff }
#define COLOR_maroon	{ 0, 0x80, 0x00, 0x00 }
#define COLOR_red	{ 0, 0xff, 0x00, 0x00 }
#define COLOR_orange	{ 0, 0xff, 0x99, 0x00 }
#define COLOR_gold	{ 0, 0xff, 0xcc, 0x00 }
#define COLOR_yellow	{ 0, 0xff, 0xff, 0x00 }
#define COLOR_green	{ 0, 0x33, 0x99, 0x00 }
#define COLOR_forestGreen { 0, 0x33, 0x99, 0x33 }
#define COLOR_seaGreen	{ 0, 0x33, 0x99, 0x66 }
#define COLOR_limeGreen	{ 0, 0x33, 0xcc, 0x33 }
#define COLOR_slateBlue	{ 0, 0x66, 0x66, 0xcc }
#define COLOR_teal	{ 0, 0x33, 0xcc, 0xcc }
#define COLOR_blue	{ 0, 0x00, 0x00, 0xff }
#define COLOR_violet	{ 0, 0x99, 0x33, 0xcc }
#define COLOR_violetRed { 0, 0xcc, 0x00, 0x99 }

extern const CustomPatternType    backPattern;
extern const CustomPatternType    emptyPattern;
extern const CustomPatternType    tablePattern;
extern const CustomPatternType    myBlackPattern;
extern IndexedColorType cardBackIndex, blackIndex, whiteIndex, packIndex,
        hintIndex;
extern void DrawTableBackground (RectangleType *r);

extern CardStackIndex	hint_src;

extern void Help (Int16 id);

extern void Rules (Int16 id);

extern void DebugS (Char*);
extern void DebugSI (Char*, PatI);

extern void NewGameForm (Boolean);
extern void StatisticsForm (void);
extern Int16 DoDialog (FormPtr, Boolean (*)(EventPtr));
extern void AdjustObject (FormPtr, UInt16,
             Int16 left_delta, Int16 top_delta,
             Int16 width_delta, Int16 height_delta);
extern void SizeFormWindow (FormPtr frmP, Int16 *dxP, Int16 *dyP);

// Song declarations

extern void PlayRandomSong (Int16 win);
extern int NSongs (void);
typedef struct _NoteType {
    unsigned char    note;
    unsigned char    duration;
} NoteType;
typedef struct _SongType {
    const NoteType *notes;
    int		    nnotes;
    int		    tempo;  // smaller is faster
} SongType;
extern const SongType *songs[];
#define RepeatEnd	0xff
#define RepeatStart	0xfe
#define Rest		0xfd

extern void (*DialogIdleFunc)(void);
extern void SelectGameEntertain (void);
extern void RestartGame (void);
extern void GameWonEntertain (void);

extern RulesType    golfGame;
extern RulesType    klondikeGame;
extern RulesType    towersGame;
extern RulesType    tabbyGame;
extern RulesType    spiderGame;
extern RulesType    montanaGame;
extern RulesType    acesGame;
extern RulesType    calcGame;
extern RulesType    yukonGame;
extern RulesType    spideretteGame;
extern RulesType    eightGame;
extern RulesType    canfieldGame;
extern RulesType    vegasGame;
extern RulesType    freecellGame;
extern RulesType    wishGame;
extern RulesType    stalactitesGame;
extern RulesType    tarantulaGame;
extern RulesType    fortyThievesGame;
extern RulesType    pictureGame;
extern RulesType    milliganGame;
extern RulesType    cotillionGame;
extern RulesType    terraceGame;
extern RulesType    bakersDozenGame;
extern RulesType    gypsyGame;

/* Increment this every time the GameType structure changes */
#define PatienceVersionNum	8
#define PatienceFileCreator	'Pati'
#define PatienceDBCreator	PatienceFileCreator
#define PatienceUndoDBType	'undo'
#define PatienceStatsDBType	'DATA'
