/*
 * $Id: pdisp.c,v 1.12 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

// debug
//#define ALLOW_ACCESS_TO_INTERNALS_OF_BITMAPS

#include "patience.h"
#include "patienceRsc.h"
#include <SonyCLIE.h>
#include <BmpGlue.h>


/*
 * Change log
 *
 * $Log: pdisp.c,v $
 * Revision 1.12  2005/01/11 17:19:00  pdclose
 * Update copyright notice
 *
 * Revision 1.11  2005/01/10 01:16:47  pdclose
 * - Animate the card moving by mapping a window to a bitmap, drawing the card,
 *   and then using the bitmap to draw the card.  That way we can have a
 *   transparent background so just the card moves, without a white square
 *   around it (3.5 and 4.x only support transparency in bitmaps).
 * - Use BMPGlueGetDimensions always instead of accessing bitmap internals when
 *   rom version < 4.0
 * - Fix some problems and compatibility issues observed on older versions of
 *   PalmOS (3.0 and 2.x)
 * - Make sure we are not in the middle of a hint when we need to draw
 *   something.
 * 
 * Revision 1.10  2004/12/28 03:49:06  pdclose
 * Reworked CardDrawRandomLoop to use multiple colors, include "You Won"
 * message and play a random song.  Renamed to GameWonEntertain.
 * 
 * Revision 1.9  2004/12/24 22:39:14  pdclose
 * Detect a problem caused when too many cards are defined.
 * 
 * Revision 1.8  2004/04/06 15:52:51  pdclose
 * Wrap common function to calculate screen coord from stack coord
 * 
 * Revision 1.7  2004/03/25 02:41:38  pdclose
 * Add support for horizontal split screen
 * 
 * Revision 1.6  2003/03/03 23:18:47  pdclose
 * Add SetHRRectangle, like SetRectangle, but 2x.
 * Add drawing routines for PalmOS 5 high-res mode.
 * Use BmpGetDimensions for accessing bitmap height/width (4.0+)
 * Add ability to draw card hints on empty cards.
 * Prevent animation for short drags.
 * 
 * Revision 1.5  2002/09/22 23:53:07  pdclose
 * Use MAX_HINT for winhandle hint buffers
 * 
 * Revision 1.4  2002/09/19 19:28:29  pdclose
 * More fussing with spacing, make HR more like regular.
 * Add support for face cards.
 * 
 * Revision 1.3  2002/07/02 21:12:23  pdclose
 * Minor tweak to suit offset.
 * 
 * Revision 1.2  2002/06/17 01:36:07  pdclose
 * Split card space into X and Y to improve spacing algorithms.
 * Tune up some of the algorithms for better spacing.
 * 
 * Revision 1.1  2002/04/23 21:54:34  pdclose
 * Initial revision
 */


static const CustomPatternType   emptyPattern = {
    0, 0, 0, 0, 0, 0, 0, 0
};

static const CustomPatternType   backPattern = {
  0xee, 0xbb, 0xdd, 0x77, 0xee, 0xbb, 0xdd, 0x77
};

static const CustomPatternType   myBlackPattern = {
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

static void CardDrawAnimate2 (PatI x, PatI y, PatI xcard, PatI ycard,
        PatI xsave, PatI ysave);
static void CardLabel (PatS x, PatS y, PatS xlimit, PatS ylimit,
        Card card, CardBool horizontal, CardState cardState) CARD_SECT;
static Boolean CardCenterBitmap (BitmapPtr bitmap,
        PatS x, PatS y, PatS xlimit, PatS ylimit,
        PatS *offsetXP, PatS *offsetYP);
static void DrawCharsOnCard (Char *buf, int len, Coord x, Coord y,
        Boolean centered, Boolean filled);
static void DrawTextInNewWindow (Char *buf, WinHandle *window,
        RectangleType *r);


static IndexedColorType
SafeWinSetForeColor(IndexedColorType colorIndex)
{
    if (romVersion >= sysVersion35)
        return WinSetForeColor(colorIndex);
    else
        return 0;
}

static IndexedColorType
SafeWinSetBackColor(IndexedColorType colorIndex)
{
    if (romVersion >= sysVersion35)
        return WinSetBackColor(colorIndex);
    else
        return 0;
}

// unless NO_MACROS is defined...
#ifndef NO_MACROS

// could also use RctSetRectangle
#define SetRectangle(r,left,top,width,height) \
    ((r)->topLeft.x = (left),	\
     (r)->topLeft.y = (top),	\
     (r)->extent.x  = (width),	\
     (r)->extent.y  = (height))

#define SetHRRectangle(r,left,top,width,height) \
    ((r)->topLeft.x = (left)*2,	\
     (r)->topLeft.y = (top)*2,	\
     (r)->extent.x  = (width)*2, \
     (r)->extent.y  = (height)*2)

// TODO could model WinScaleRectangle and detect hrMode, then use always
#define HRRectangle(r)	\
    ((r)->topLeft.x *= 2,	\
     (r)->topLeft.y *= 2,	\
     (r)->extent.x *= 2,	\
     (r)->extent.y *= 2)

#define CardExtent(x,y,r) SetRectangle((r),(x),(y),cardWidth,cardHeight)

#define HRDrawBitmap(bitmap,x,y) \
            ((romVersion >= sysVersion35)? \
                HRWinPaintBitmap(hrRefNum, (bitmap), (x), (y)): \
                HRWinDrawBitmap(hrRefNum, (bitmap), (x), (y)))

#define DrawBitmap(bitmap,x,y) \
            ((romVersion >= sysVersion35)? \
                WinPaintBitmap((bitmap), (x), (y)): \
                WinDrawBitmap((bitmap), (x), (y)))

#define CopyToScreen(srcWin,r,destX,destY) \
        CopyRectangle((srcWin),WinGetDrawWindow(),(r),(destX),(destY))

#define CopyFromScreen(destWin,r,destX,destY) \
        CopyRectangle(WinGetDrawWindow(),(destWin),(r),(destX),(destY))

#define CoordXtoScreen(x) \
        ((x) * cardWidth + cardOriginX)

#define CoordYtoScreen(y) \
        ((y) * cardHeight + cardOriginY)

#define GetXLimit(sp) \
        (form_width)

#define GetYLimit(sp) \
        (column[(sp)->pos.x].hasArb? \
            column[(sp)->pos.x].height[(sp)->pos.y] : form_height)

#else
    
static void
SetRectangle (RectangleType *rP,
	Coord left, Coord top, Coord width, Coord height)
{
    rP->topLeft.x = left;
    rP->topLeft.y = top;
    rP->extent.x = width;
    rP->extent.y = height;
}

static void
SetHRRectangle (RectangleType *rP,
	Coord left, Coord top, Coord width, Coord height)
{
    rP->topLeft.x = left*2;
    rP->topLeft.y = top*2;
    rP->extent.x = width*2;
    rP->extent.y = height*2;
}

static void
HRRectangle(RectangleType *rP)
{
    rP->topLeft.x *= 2;
    rP->topLeft.y *= 2;
    rP->extent.x *= 2;
    rP->extent.y *= 2;
}

static void
CardExtent (PatI x, PatI y, RectangleType *r)
{
    r->topLeft.x = x;
    r->topLeft.y = y;
    r->extent.x = cardWidth;
    r->extent.y = cardHeight;
}

static void
HRDrawBitmap(BitmapType *bitmap, Coord x, Coord y)
{
    if (romVersion >= sysVersion35)
        HRWinPaintBitmap (hrRefNum, bitmap, x, y);
    else
        HRWinDrawBitmap (hrRefNum, bitmap, x, y);
}

static void
DrawBitmap(BitmapType *bitmap, Coord x, Coord y)
{
    if (romVersion >= sysVersion35)
        WinPaintBitmap (bitmap, x, y);
    else
        WinDrawBitmap (bitmap, x, y);
}

static PatS
CoordXtoScreen (PatS x)
{
    return ((x) * cardWidth + cardOriginX);
}

static PatS
CoordYtoScreen (PatS y)
{
    return ((y) * cardHeight + cardOriginY);
}

static PatS
GetXLimit (CardStackType *sp)
{
    return form_width;
}

static PatS
GetYLimit (CardStackType *sp)
{
    if (column[sp->pos.x].hasArb)
        return column[sp->pos.x].height[sp->pos.y];
    else
        return form_height;
}

#endif /*NO_MACROS*/

// like WinCreateOffscreenWindow
static WinHandle
CreateWindow(Coord width, Coord height, UInt16 *errorP)
{
    WinHandle       saveWin;

    switch (hrMode)
    {
      case SonyHR:
        saveWin = HRWinCreateOffscreenWindow (hrRefNum,
                width*2, height*2, screenFormat, errorP);
        break;
      case PalmOS5HR:
        saveWin = WinCreateOffscreenWindow (
                width*2, height*2, nativeFormat, errorP);
        break;
      default:
        saveWin = WinCreateOffscreenWindow (
                width, height, screenFormat, errorP);
        break;
    }

    return saveWin;
}

// somewhat like WinCreateBitmapWindow
// TODO: Likely does not handle Palm OS 5 properly
static WinHandle
CreateBitmapWindow (Coord width, Coord height, UInt8 depth,
        BitmapType **bitmapP)
{
    BitmapType *bitmap;
    UInt16	error;

    if (hrMode==SonyHR)
        bitmap = HRBmpCreate (hrRefNum, width*2, height*2, depth, NULL, &error);
    else
        bitmap = BmpCreate (width, height, depth, NULL, &error);

    *bitmapP = bitmap;
    if (bitmap == NULL)
        return NULL;

    BmpGlueSetTransparentValue (bitmap, 0);
    return WinCreateBitmapWindow (bitmap, &error);
}

// like WinCopyRectangle
static void
CopyRectangle2(WinHandle srcWin, WinHandle destWin,
        const RectangleType *srcRect, Coord destX, Coord destY,
        WinDrawOperation mode)
{
    RectangleType r;

    switch (hrMode)
    {
      case SonyHR:
        RctCopyRectangle(srcRect, &r);
        HRRectangle (&r);
        HRWinCopyRectangle (hrRefNum,
                srcWin, destWin, &r, destX*2, destY*2, mode);
        break;
      case PalmOS5HR:
        RctCopyRectangle(srcRect, &r);
        HRRectangle (&r);
        WinCopyRectangle (srcWin, destWin, &r, destX*2, destY*2, mode);
        break;
      default:
        WinCopyRectangle (srcWin, destWin, srcRect, destX, destY, mode);
        break;
    }
}

static void
CopyRectangle(WinHandle srcWin, WinHandle destWin,
        const RectangleType *srcRect, Coord destX, Coord destY)
{
    CopyRectangle2(srcWin, destWin, srcRect, destX, destY, winPaint);
}

static void
CardRect (PatI x, PatI y, RectangleType *r)
{
    SetRectangle(r, x+cardSpaceX/2, y+cardSpaceY/2,
            cardWidth-cardSpaceX, cardHeight-cardSpaceY);
}

static PatI
StackOff (CardStackIndex stack, PatI pos, CardState *cardStateP)
{
    CardStackType   *sp = CardStackPtr (stack);
    PatI	    off;
    CardIndex	    card;
    CardType	    *cp = NULL;

    off = 0;
    for (card = sp->first; pos && card != NullIndex; card = cp->next)
    {
	cp = CardPtr (card);
	if (cp->show != CardShowNone)
	{
	    switch (cp->show)
	    {
	    case CardShowBack:
		off += sp->back_off;
		break;
	    case CardShowPack:
		off += sp->pack_off;
		break;
	    case CardShowFace:
		off += sp->face_off;
		break;
	    }
	    pos--;
	}
    }
    if (cardStateP != NULL)
    {
        if (card != NullIndex)
        {
            cp = CardPtr (card);
            if (cp->show == CardShowPack)
                *cardStateP = cardStatePacked;
            else if (card == sp->last || cp->show == CardShowNone)
                *cardStateP = cardStateTop;
            else
                *cardStateP = cardStateOther;
        }
        else
            *cardStateP = cardStateNone;
    }
    return off;
}

static void
StackExtent (CardStackIndex stack, RectangleType *r)
{
    CardStackType   *sp = CardStackPtr (stack);
    PatI	    off;

    CardExtent (CoordXtoScreen(sp->pos.x), CoordYtoScreen(sp->pos.y), r);

    off = sp->face_up * sp->face_off + sp->pack_up * sp->pack_off + 
	  sp->back_up * sp->back_off;

    if (sp->horizontal) {
        r->extent.x += off;
        if (r->extent.x > GetXLimit(sp))
            r->extent.x = GetXLimit(sp);
    }
    else {
        r->extent.y += off;
        if (r->extent.y > GetYLimit(sp))
            r->extent.y = GetYLimit(sp) + 2;  // compensate for -2 below
    }

    // sets the extent more accurately on top and bottom
    // TODO seems to work, but not sure why or what else might be needed
    r->topLeft.y++;
    r->extent.y -= 2;
}

/*
 * This routine now fills the outline as well.
 */
static void
CardOutlineDraw (PatI x, PatI y, CardBool empty)
{
    RectangleType   r;
    PatI	    off_x, off_y, space_x, space_y;

    space_x = cardSpaceX;
    if (empty && space_x < 2)
        space_x = 2;
    off_x = space_x / 2;

    space_y = cardSpaceY;
    if (empty && space_y < 2)
        space_y = 2;
    off_y = space_y / 2;

    switch (hrMode)
    {
      case SonyHR:
        // Adjust a little to compensate for new extent setting in
        // StackExtent
        // TODO work in progress
        //off_y++;
        //space_y += 2;
        //off_x++;
        //space_x += 2;
        //SetRectangle (&r, x*2 + off_x, y*2 + off_y,
        //        cardWidth*2 - space_x, cardHeight*2 - space_y);
        SetHRRectangle (&r, x + off_x, y + off_y,
                cardWidth - space_x, cardHeight - space_y);
        if (!empty)
            HRWinEraseRectangle (hrRefNum, &r, 4);
        HRWinDrawRectangleFrame (hrRefNum, roundFrame, &r);
        break;
      case PalmOS5HR:
        SetHRRectangle (&r, x + off_x, y + off_y,
                cardWidth - space_x, cardHeight - space_y);
        if (!empty)
            WinEraseRectangle (&r, cardRadius);
        WinDrawRectangleFrame ((cardRadius << 8) | 1, &r);
        break;
      default:
        SetRectangle (&r, x + off_x, y + off_y,
                cardWidth - space_x, cardHeight - space_y);

        if (!empty)
            WinEraseRectangle (&r, cardRadius);
        WinDrawRectangleFrame ((cardRadius << 8) | 1, &r);
        break;
    }
}

static void
CardErase (PatI x, PatI y, Boolean table)
{
    RectangleType   r;

    CardExtent (x, y, &r);
    // sets the extent more accurately on top and bottom
    // TODO seems to work, but not sure why or what else might be needed
    r.topLeft.y++;
    r.extent.y -= 2;

    if (hrMode==PalmOS5HR)
        HRRectangle(&r);

    if (table)
        DrawTableBackground (&r);
    else
        WinEraseRectangle (&r, 0);
}

#define card_to_suit_id(card)	(bitmapSuitBase + (Suit(card) >> 4))
#define card_to_rank_id(card)	(bitmapRankBase + (Rank(card) == ACE ? \
						   1 : Rank(card)))
typedef struct _CardImage {
    MemHandle	hand;
    BitmapPtr	bitmap;
} CardImage;

// A, 1-10, J, Q, K
static CardImage    CardRankBitmaps[14];
// Narrow 10 for narrow cards
static CardImage    CardRankBitmap10Narrow;
// upper corner suits: club, diamond, heart, spade
static CardImage    CardSuitBitmaps[4];
// large (centered) suits
static CardImage    CardLgSuitBitmaps[4];
// First dim: Jack, Queen, King, second dim is club, diamond, heart, spade
static CardImage    CardLgFaceBitmaps[3][4];

static void
CardImageInit (CardImage *image, Int16 res, Boolean checkSize)
{
    Coord width, height;
    if ((image->hand = DmGetResource (bitmapRsc, res)))
    {
	image->bitmap = MemHandleLock (image->hand);
	if (image->bitmap && checkSize)
	{
            BmpGlueGetDimensions (image->bitmap, &width, &height, NULL);
            if (width > cardLabelWidth)
                cardLabelWidth = width;
            if (height > cardLabelHeight)
                cardLabelHeight = height;
	}
    }
}

static void
CardImageFini (CardImage *image)
{
    if (image->hand)
    {
	if (image->bitmap)
	{
	    MemHandleUnlock (image->hand);
	    image->bitmap = 0;
	}
	DmReleaseResource (image->hand);
	image->hand = 0;
    }
}

void
CardLabelInit (void)
{
    int	    rank;

    if (hrMode)
    {
        for (rank = 1; rank <= KING; rank++)
            CardImageInit (&CardRankBitmaps[rank], bitmapHRRankBase + rank,
                    true);
        CardImageInit (&CardSuitBitmaps[CLUB>>4], bitmapHRClubSuit, true);
        CardImageInit (&CardSuitBitmaps[DIAMOND>>4], bitmapHRDiamondSuit, true);
        CardImageInit (&CardSuitBitmaps[HEART>>4], bitmapHRHeartSuit, true);
        CardImageInit (&CardSuitBitmaps[SPADE>>4], bitmapHRSpadeSuit, true);
        CardImageInit (&CardLgSuitBitmaps[CLUB>>4], bitmapHRClubSuitLg,
                false);
        CardImageInit (&CardLgSuitBitmaps[DIAMOND>>4], bitmapHRDiamondSuitLg,
                false);
        CardImageInit (&CardLgSuitBitmaps[HEART>>4], bitmapHRHeartSuitLg,
                false);
        CardImageInit (&CardLgSuitBitmaps[SPADE>>4], bitmapHRSpadeSuitLg,
                false);
        // not needed for high-res, use regular 10
        CardImageInit (&CardRankBitmap10Narrow, bitmapHRRank10, true);

        // Face card images, if present
        CardImageInit (&CardLgFaceBitmaps[0][CLUB>>4], bitmapHRJackClub,
                false);
        CardImageInit (&CardLgFaceBitmaps[0][DIAMOND>>4], bitmapHRJackDiamond,
                false);
        CardImageInit (&CardLgFaceBitmaps[0][HEART>>4], bitmapHRJackHeart,
                false);
        CardImageInit (&CardLgFaceBitmaps[0][SPADE>>4], bitmapHRJackSpade,
                false);
        CardImageInit (&CardLgFaceBitmaps[1][CLUB>>4], bitmapHRQueenClub,
                false);
        CardImageInit (&CardLgFaceBitmaps[1][DIAMOND>>4], bitmapHRQueenDiamond,
                false);
        CardImageInit (&CardLgFaceBitmaps[1][HEART>>4], bitmapHRQueenHeart,
                false);
        CardImageInit (&CardLgFaceBitmaps[1][SPADE>>4], bitmapHRQueenSpade,
                false);
        CardImageInit (&CardLgFaceBitmaps[2][CLUB>>4], bitmapHRKingClub,
                false);
        CardImageInit (&CardLgFaceBitmaps[2][DIAMOND>>4], bitmapHRKingDiamond,
                false);
        CardImageInit (&CardLgFaceBitmaps[2][HEART>>4], bitmapHRKingHeart,
                false);
        CardImageInit (&CardLgFaceBitmaps[2][SPADE>>4], bitmapHRKingSpade,
                false);
    }
    else if (qvgaMode)
    {
        for (rank = 1; rank <= KING; rank++)
            CardImageInit (&CardRankBitmaps[rank], bitmapQVGARankBase + rank,
                    true);
        CardImageInit (&CardRankBitmap10Narrow, bitmapQVGARank10N, true);

        // for QVGA, large = small 2x icon
        CardImageInit (&CardSuitBitmaps[CLUB>>4], bitmapQVGAClubSuit, true);
        CardImageInit (&CardSuitBitmaps[DIAMOND>>4], bitmapQVGADiamondSuit,
                true);
        CardImageInit (&CardSuitBitmaps[HEART>>4], bitmapQVGAHeartSuit, true);
        CardImageInit (&CardSuitBitmaps[SPADE>>4], bitmapQVGASpadeSuit, true);

        CardImageInit (&CardLgSuitBitmaps[CLUB>>4], bitmapHRClubSuit, false);
        CardImageInit (&CardLgSuitBitmaps[DIAMOND>>4], bitmapHRDiamondSuit,
                false);
        CardImageInit (&CardLgSuitBitmaps[HEART>>4], bitmapHRHeartSuit, false);
        CardImageInit (&CardLgSuitBitmaps[SPADE>>4], bitmapHRSpadeSuit, false);

        // Face card images, if present
        CardImageInit (&CardLgFaceBitmaps[0][CLUB>>4], bitmapQVGAJackClub,
                false);
        CardImageInit (&CardLgFaceBitmaps[0][DIAMOND>>4],
                bitmapQVGAJackDiamond, false);
        CardImageInit (&CardLgFaceBitmaps[0][HEART>>4], bitmapQVGAJackHeart,
                false);
        CardImageInit (&CardLgFaceBitmaps[0][SPADE>>4], bitmapQVGAJackSpade,
                false);
        CardImageInit (&CardLgFaceBitmaps[1][CLUB>>4], bitmapQVGAQueenClub,
                false);
        CardImageInit (&CardLgFaceBitmaps[1][DIAMOND>>4],
                bitmapQVGAQueenDiamond, false);
        CardImageInit (&CardLgFaceBitmaps[1][HEART>>4], bitmapQVGAQueenHeart,
                false);
        CardImageInit (&CardLgFaceBitmaps[1][SPADE>>4], bitmapQVGAQueenSpade,
                false);
        CardImageInit (&CardLgFaceBitmaps[2][CLUB>>4], bitmapQVGAKingClub,
                false);
        CardImageInit (&CardLgFaceBitmaps[2][DIAMOND>>4],
                bitmapQVGAKingDiamond, false);
        CardImageInit (&CardLgFaceBitmaps[2][HEART>>4], bitmapQVGAKingHeart,
                false);
        CardImageInit (&CardLgFaceBitmaps[2][SPADE>>4], bitmapQVGAKingSpade,
                false);
    }
    else
    {
        for (rank = 1; rank <= KING; rank++)
            CardImageInit (&CardRankBitmaps[rank], bitmapRankBase + rank, true);
        CardImageInit (&CardSuitBitmaps[CLUB>>4], bitmapClubSuit, true);
        CardImageInit (&CardSuitBitmaps[DIAMOND>>4], bitmapDiamondSuit, true);
        CardImageInit (&CardSuitBitmaps[HEART>>4], bitmapHeartSuit, true);
        CardImageInit (&CardSuitBitmaps[SPADE>>4], bitmapSpadeSuit, true);
        CardImageInit (&CardLgSuitBitmaps[CLUB>>4], bitmapClubSuitLg, false);
        CardImageInit (&CardLgSuitBitmaps[DIAMOND>>4], bitmapDiamondSuitLg,
                false);
        CardImageInit (&CardLgSuitBitmaps[HEART>>4], bitmapHeartSuitLg, false);
        CardImageInit (&CardLgSuitBitmaps[SPADE>>4], bitmapSpadeSuitLg, false);
        CardImageInit (&CardRankBitmap10Narrow, bitmapRank10N, true);
    }
}

void
CardLabelFini (void)
{
    int	    rank, suit;

    for (rank = 1; rank <= KING; rank++)
	CardImageFini (&CardRankBitmaps[rank]);
    for (suit = 0; suit < 4; suit++)
	CardImageFini (&CardSuitBitmaps[suit]);
    for (suit = 0; suit < 4; suit++)
	CardImageFini (&CardLgSuitBitmaps[suit]);
    CardImageFini (&CardRankBitmap10Narrow);

    // Face card images, if present
    for (suit = 0; suit < 4; suit++)
    {
        CardImageFini (&CardLgFaceBitmaps[0][suit]);
        CardImageFini (&CardLgFaceBitmaps[1][suit]);
        CardImageFini (&CardLgFaceBitmaps[2][suit]);
    }
    cardLabelWidth = 0;
    cardLabelHeight = 0;
}

static void
CardLabel (PatS x, PatS y, PatS xlimit, PatS ylimit,
        Card card, CardBool horizontal, CardState cardState)
{
    BitmapPtr	rank_bitmap = NULL;
    BitmapPtr	suit_bitmap = NULL;
    BitmapPtr	face_bitmap = NULL;
    Boolean	center_top_suit = false;
    Boolean	face_card_image = false;

    Coord	card_width = 0;
    Coord	rank_width = 0;
    Coord	rank_height = 0;
    Coord	suit_width = 0;
    Coord	card_height = 0;

    PatS	rank_offset_x = 0;
    PatS	rank_offset_y = 0;
    PatS	suit_offset_x = 0;
    PatS	suit_offset_y = 0;
    PatS	face_offset_x = 0;
    PatS	face_offset_y = 0;

    /*
     * white portion of card is cardWidth-cardSpaceX wide
     * 10 cards, cardWidth=16, cardSpaceX=1                     (40 Thieves)
     * 9 cards, cardWidth=18, cardSpaceX=1 (see SetSizes)       (Miss Milligan)
     * 8 cards, cardWidth=20, cardSpaceX=2                      (Stalactites)
     * <=7 cards, cardWidth=22, cardSpaceX=4                    (Klondike)
     *
     * See SetSizes in patience.c for algorithm
     */

    x += (cardSpaceX/2 + 1);
    y += (cardSpaceY/2 + 1);

    // width/height of white part of card
    card_width = cardWidth - cardSpaceX;
    card_height = cardHeight - cardSpaceY;
    if (hrMode)
    {
        card_width *= 2;
        card_height *= 2;
    }

    rank_bitmap = CardRankBitmaps[Rank(card) == ACE ? 1 : Rank(card)].bitmap;
    suit_bitmap = CardSuitBitmaps[Suit(card) >> 4].bitmap;

    // use narrow 10 when necessary
    if (Rank(card) == 10 && horizontal && cardWidth < maxCardWidth)
        rank_bitmap = CardRankBitmap10Narrow.bitmap;

    if (rank_bitmap != 0)
        BmpGlueGetDimensions (rank_bitmap, &rank_width, &rank_height, NULL);
    if (suit_bitmap != 0)
        BmpGlueGetDimensions (suit_bitmap, &suit_width, NULL, NULL);

    if (horizontal)
    {
	/* rank_offset_x = 0; */
	/* rank_offset_y = 0; */
	suit_offset_x = rank_width + 1;
        if (hrMode)
            suit_offset_x++;
	/* suit_offset_y = 0; */
    }
    else
    {
	if (rank_width > suit_width)
	{
	    /* rank_offset_x = 0; */
	    suit_offset_x = (rank_width - suit_width) / 2;
	}
	else
	{
	    rank_offset_x = (suit_width - rank_width) / 2;
	    /* suit_offset_x = 0; */
	}
	/* rank_offset_y = 0; */
	suit_offset_y = yOffset;

        // add back the 1/2 cardLabelHeight we took off in SetSizes()
        if (hrMode)
            suit_offset_y += cardLabelHeight/2;
    }

    center_top_suit = (game.centerTopSuit && cardState == cardStateTop);

    if (center_top_suit)
    {
        BitmapPtr tmp_bitmap;

        // special - face card images
        if (Rank(card) >= JACK && Rank(card) <= KING) 
        {
            tmp_bitmap = CardLgFaceBitmaps[Rank(card) - JACK][Suit(card) >> 4].
                    bitmap;
            if (tmp_bitmap && CardCenterBitmap(tmp_bitmap,
                    x, y, xlimit, ylimit, &face_offset_x, &face_offset_y))
            {
                face_card_image = true;
                face_bitmap = tmp_bitmap;
                // adjust for the +1 added to x earlier
                face_offset_x -= (hrMode? 2 : 1);
                face_offset_y += rank_height / 2;
                center_top_suit = false;
                // Avoid overlapping the face card and small suit
                // (seen in montana HandEra rotated)
                // Note - a better solution would be to force the suit up
                // using the horizontal alignment.
                if (!horizontal)
                {
                    //suit_bitmap = NULL;
                    horizontal = true;
                    rank_offset_x = 0;
                    rank_offset_y = 0;
                    suit_offset_x = rank_width + 1;
                    if (hrMode)
                        suit_offset_x++;
                    suit_offset_y = 0;
                }
            }
        }

        // centered large suit
        tmp_bitmap = CardLgSuitBitmaps[Suit(card) >> 4].bitmap;
        if (center_top_suit && CardCenterBitmap(tmp_bitmap,
                x, y, xlimit, ylimit, &suit_offset_x, &suit_offset_y))
        {
            suit_bitmap = tmp_bitmap;
            BmpGlueGetDimensions (suit_bitmap, &suit_width, NULL, NULL);
            // adjust for the +1 added to x earlier
            suit_offset_x -= (hrMode? 2 : 1);

            // undo possible narrow 10
            if (Rank(card) == 10)
                rank_bitmap = CardRankBitmaps[10].bitmap;
        }
        else
        {
            center_top_suit = false;
        }
    }

    // Not centered top card, squish if necessary
    if (!center_top_suit && horizontal &&
         ( (hrMode && suit_offset_x + suit_width + 4 >= cardWidth*2) ||
           (!hrMode && suit_offset_x + suit_width + 2 >= cardWidth) )
       )
    {
        rank_offset_x--;
        suit_offset_x -= (hrMode? 2 : 1);
    }

    // Need to paint using overlay to get transparent backgrounds in
    // greyscale icons
    if (romVersion >= sysVersion35 && screen_depth < 8)
    {
        WinPushDrawState ();
        WinSetDrawMode (winOverlay);
    }

    switch (hrMode)
    {
      case SonyHR:
        if (rank_bitmap)
            HRDrawBitmap (rank_bitmap,
                    x*2 + rank_offset_x, y*2 + rank_offset_y);

        if (suit_bitmap)
            HRDrawBitmap (suit_bitmap,
                    x*2 + suit_offset_x, y*2 + suit_offset_y);

        if (face_bitmap)
            HRDrawBitmap (face_bitmap,
                    x*2 + face_offset_x, y*2 + face_offset_y);
        break;
      case PalmOS5HR:
        if (rank_bitmap)
            DrawBitmap (rank_bitmap, x*2 + rank_offset_x, y*2 + rank_offset_y);

        if (suit_bitmap)
            DrawBitmap (suit_bitmap, x*2 + suit_offset_x, y*2 + suit_offset_y);

        if (face_bitmap)
            DrawBitmap (face_bitmap, x*2 + face_offset_x, y*2 + face_offset_y);
        break;
      default:
        if (rank_bitmap)
            DrawBitmap (rank_bitmap, x + rank_offset_x, y + rank_offset_y);

        if (suit_bitmap)
            DrawBitmap (suit_bitmap, x + suit_offset_x, y + suit_offset_y);

        if (face_bitmap)
            DrawBitmap (face_bitmap, x + face_offset_x, y + face_offset_y);
        break;
    }

    if (romVersion >= sysVersion35 && screen_depth < 8)
        WinPopDrawState ();
}

static Boolean
CardCenterBitmap(BitmapPtr bitmap, PatS x, PatS y, PatS xlimit, PatS ylimit,
        PatS *offsetXP, PatS *offsetYP)
{
    Boolean     center = true;
    short	offset_x;
    short	offset_y;

    Coord       bitmap_width;
    Coord       bitmap_height;
    
    if (bitmap == NULL)
        return false;

    BmpGlueGetDimensions (bitmap, &bitmap_width, &bitmap_height, NULL);

    if (hrMode)
    {
        // cardWidth - cardSpaceX is the width of the white part of the card
        offset_x = ((cardWidth - cardSpaceX)*2 - bitmap_width + 1) / 2;
        offset_y = ((cardHeight - cardSpaceY)*2 - bitmap_height + 1) / 2;
        // Check if there's room for a centered bitmap
        if (offset_x < 0 || offset_y < 0 ||
                y*2 + offset_y + bitmap_height > ylimit*2)
            center = false;
    }
    else
    {
        // cardWidth - cardSpaceX is the width of the white part of the card
        offset_x = (cardWidth - cardSpaceX - bitmap_width + 1) / 2;
        offset_y = (cardHeight - cardSpaceY - bitmap_height + 1) / 2;
        // Check if there's room for a centered bitmap
        if (offset_x < 0 || offset_y < 0 ||
                y + offset_y + bitmap_height > ylimit)
            center = false;
    }

    if (center)
    {
        if (offsetXP)
            *offsetXP = offset_x;
        if (offsetYP)
            *offsetYP = offset_y;
    }

    return center;
}

static void
CardBack (PatI x, PatI y)
{
    RectangleType   r;
    IndexedColorType currForeColor = 0;

    //TODO - Some way to detect the "pack" and draw that using popupFrame?
    CardRect (x, y, &r);
    // make the pattern smaller than the card for a white border
    r.topLeft.x++;
    r.topLeft.y++;
    r.extent.x -= 2;
    r.extent.y -= 2;
    if (hrMode==PalmOS5HR)
        HRRectangle (&r);
    currForeColor = SafeWinSetForeColor(cardBackIndex);
    WinSetPattern (&backPattern);
    WinFillRectangle (&r, cardRadius);
    SafeWinSetForeColor(currForeColor);
    WinSetPattern (&emptyPattern);
}

void
CardDrawPos (PatS x, PatS y, PatS xlimit, PatS ylimit,
        Card card, CardBool horizontal, CardState cardState)
{
    IndexedColorType currBackColor = 0;

    switch (Suit (card)) {
    case CLUB: case DIAMOND: case HEART: case SPADE:
        if (game.forceBW)
            CardErase (x, y, false);
        if (game.alwaysPack && cardState == cardStatePacked)
            currBackColor = SafeWinSetBackColor(packIndex);
	CardOutlineDraw (x, y, false);
	CardLabel (x, y, xlimit, ylimit, card, !horizontal, cardState);
        if (game.alwaysPack && cardState == cardStatePacked)
            SafeWinSetBackColor(currBackColor);
	break;
    case BACK:
        if (game.forceBW)
            CardErase (x, y, false);
	CardOutlineDraw (x, y, false);
	CardBack (x, y);
	break;
    case EMPTY:
        if (game.forceBW)
            CardErase (x, y, true);
	CardOutlineDraw (x, y, true);
        // card hints
        if (Rank (card) != 0)
        {
            short rank = Rank (card);
            short len = 1;
            char buf[2];
            switch (rank)
            {
              case 1:
              case ACE:
                *buf = 'A'; break;
              case 10:
                *buf = '1'; buf[1] = '0'; len = 2; break;
              case JACK:
                *buf = 'J'; break;
              case QUEEN:
                *buf = 'Q'; break;
              case KING:
                *buf = 'K'; break;
              default:
                *buf = '0' + rank; break;
            }
            DrawCharsOnCard(buf, len, x, y, true, false);
        }
	break;
    case NONE:
	CardErase (x, y, true);
	break;
    case BLANK:
        if (game.forceBW)
            CardErase (x, y, false);
	CardOutlineDraw (x, y, false);
	break;
    }
}

static void
CardDraw (CardStackIndex stack, PatI pos, Card card)
{
    CardStackType   *sp = CardStackPtr (stack);
    PatI    	    x, y;
    CardState	    cardState;

    x = CoordXtoScreen(sp->pos.x);
    y = CoordYtoScreen(sp->pos.y);
    if (sp->horizontal)
	x += StackOff (stack, pos, &cardState);
    else
	y += StackOff (stack, pos, &cardState);
    CardDrawPos (x, y, GetXLimit(sp), GetYLimit(sp), card,
            sp->horizontal, cardState);
}

// "fireworks"
void
GameWonEntertain (void)
{
    RectangleType   clip, oldClip;
    Coord           x, y;
    IndexedColorType colors[6];
    UInt16          colorIdx = 0;
    WinHandle       textWin[6];
    RectangleType   textRect;
    Coord           textX, textY;

    int		    n = 0;
    int		    repeating = 0;
    UInt32          nextNoteTick = 0;
    const NoteType *note;
    const SongType *song;
    SndCommandType  cmd;

    // setup

    song = songs[SysRandom(0) % NSongs ()];

    if (romVersion >= sysVersion35)
    {
        const RGBColorType red    = COLOR_red;
        const RGBColorType orange = COLOR_orange;
        const RGBColorType green  = COLOR_limeGreen;
        const RGBColorType blue   = COLOR_blue;
        const RGBColorType violet = COLOR_violetRed;
        colors[0] = blackIndex;
        colors[1] = WinRGBToIndex(&red);
        colors[2] = WinRGBToIndex(&orange);
        colors[3] = WinRGBToIndex(&green);
        colors[4] = WinRGBToIndex(&blue);
        colors[5] = WinRGBToIndex(&violet);

        WinPushDrawState ();
        WinSetForeColor (blackIndex);
        WinSetBackColor (whiteIndex);
        if (hrMode==PalmOS5HR)
            WinSetCoordinateSystem (kCoordinatesDouble);
    }
    for (colorIdx=0; colorIdx < 6; colorIdx++) {
        if (romVersion >= sysVersion35) {
            WinSetForeColor (colors[colorIdx]);
            WinSetTextColor (colors[colorIdx]);
        }
        DrawTextInNewWindow ("You  Won!", &textWin[colorIdx], &textRect);
    }
    if (hrMode==PalmOS5HR || hrMode==SonyHR) {
        textX = form_width - textRect.extent.x / 2;
        textY = form_height - textRect.extent.y / 2;
        // compensate for doubling in CopyToScreen
        textX /= 2; textY /= 2;
    } else {
        textX = (form_width - textRect.extent.x) / 2;
        textY = (form_height - textRect.extent.y) / 2;
    }

    WinGetClip (&oldClip);
    SetRectangle(&clip, 0, 0, form_width, form_height);
    if (hrMode==PalmOS5HR)
        HRRectangle(&clip);
    WinSetClip (&clip);

    // enter drawing / playing loop

    while (true)
    {
        EventType    event;
        EvtGetEvent (&event, 0);
        if (event.eType != nilEvent)
        {
            break;
        }

        // display a random card

        x = (SysRandom(0) % (form_width + cardWidth));
        y = (SysRandom(0) % (form_height + cardHeight));
        // offset cards 1/2 width and height to fill window more evenly
        if (romVersion >= sysVersion33)
        {
            // BUG noticed in PalmOS 3.0 that this caused an error when the
            // label bitmap was drawn completely offscreen (-10,-8)
            x = (SysRandom(0) % (form_width + cardWidth)) - cardWidth/2;
            y = (SysRandom(0) % (form_height + cardHeight)) - cardHeight/2;
        }
        if (romVersion >= sysVersion35)
            WinSetForeColor (colors[colorIdx++ % 6]);
        CardDrawPos (x, y, form_width, form_height,
                MakeCard (0, (SysRandom(0) % 4) << 4, (SysRandom(0) % 13)+1),
                false, cardStateTop);
        //TODO only redraw text if the card overlapped it?
        CopyToScreen (textWin[(colorIdx/15) % 6], &textRect, textX, textY);

        // play a note

        if (romVersion >= sysVersion30 &&
                game.volume > 0 && n < song->nnotes && 
                TimGetTicks() > nextNoteTick)
        {
            note = song->notes + n;
            switch (note->note) {
            case RepeatEnd:
                if (repeating) {
                    repeating = 0;
                    n++;
                } else {
                    while (n >= 0) {
                        if (song->notes[n].note == RepeatStart)
                            break;
                        n--;
                    }
                    repeating = 1;
                }
                break;
            case RepeatStart:
                n++;
                break;
            default:
                cmd.param2 = note->duration * song->tempo;
                nextNoteTick = TimGetTicks() + cmd.param2 *
                        sysTicksPerSecond / 1000;
                if (note->note != Rest)
                {
                    cmd.cmd = sndCmdNoteOn;
                    //cmd.param1 = Freq(note->note);
                    // 45 is the midi note for the A below middle C (60)
                    cmd.param1 = 45 + (note->note & 0x0f) +
                            12 * (6-((note->note & 0xf0)>>4));
                    cmd.param3 = game.volume;
                    SndDoCmd (0, &cmd, True);
                }
                n++;
                break;
            }
        }

        SysTaskDelay (1);
    }

    for (colorIdx=0; colorIdx < 6; colorIdx++)
        WinDeleteWindow (textWin[colorIdx], false);
    WinSetClip (&oldClip);
    if (romVersion >= sysVersion35)
        WinPopDrawState ();
    FrmUpdateForm (mainForm, frmRedrawUpdateCode);
}

static void
DrawTextInNewWindow (Char *buf, WinHandle *window, RectangleType *r)
{
    Coord           x, y;
    int		    w, h, len;
    FontID          prevFont = -1;
    RectangleType   rFrame;
    WinHandle       draw;
    UInt16          error;

    if (romVersion >= sysVersion30) {
        if (hrMode==SonyHR)
            prevFont = HRFntSetFont (hrRefNum, hrLargeBoldFont);
        else
            prevFont = FntSetFont (largeBoldFont);
    } else
        prevFont = FntSetFont (largeFont);
    len = StrLen (buf);
    w = FntLineWidth (buf, len);
    h = FntLineHeight ();
    if (hrMode==PalmOS5HR || hrMode==SonyHR) {
        x = 8; y = 4;
        SetRectangle (r, 0, 0, w + x*2, h + y*2);
        // Frames are drawn outside the rectangle
        SetRectangle (&rFrame, 2, 2, r->extent.x-4, r->extent.y-4);
    } else {
        x = 4; y = 2;
        SetRectangle (r, 0, 0, w + x*2, h + y*2);
        SetRectangle (&rFrame, 1, 1, r->extent.x-2, r->extent.y-2);
    }

    if (hrMode==SonyHR) {
        *window = HRWinCreateOffscreenWindow (hrRefNum,
                r->extent.x, r->extent.y, screenFormat, &error);
    }
    else if (hrMode==PalmOS5HR)
        *window = WinCreateOffscreenWindow (
                r->extent.x, r->extent.y, nativeFormat, &error);
    else
        *window = WinCreateOffscreenWindow (
                r->extent.x, r->extent.y, screenFormat, &error);
    // if (!*window) ???
    draw = WinGetDrawWindow ();
    WinSetDrawWindow (*window);
    if (hrMode==PalmOS5HR)
        WinSetCoordinateSystem (kCoordinatesDouble);

    if (hrMode==SonyHR) {
        HRWinEraseRectangle (hrRefNum, r, 0);
        HRWinDrawRectangleFrame (hrRefNum, 2, &rFrame);
        HRWinDrawChars (hrRefNum, buf, len, x, y);
        HRFntSetFont(hrRefNum, prevFont);
    } else {
        WinEraseRectangle (r, 0);
        WinDrawRectangleFrame (1, &rFrame);
        WinDrawChars (buf, len, x, y);
        FntSetFont(prevFont);
    }
    WinSetDrawWindow (draw);
}

static void
CardEraseStack (CardStackIndex stack)
{
    RectangleType   r;

    StackExtent (stack, &r);
    if (hrMode==PalmOS5HR)
        HRRectangle (&r);
    DrawTableBackground (&r);
}

static void
SetStackClip (CardStackType *sp, RectangleType *oldClip)
{
    RectangleType clip;

    WinGetClip (oldClip);

    CardExtent (CoordXtoScreen(sp->pos.x), CoordYtoScreen(sp->pos.y), &clip);

    if (sp->horizontal)
        clip.extent.x = GetXLimit(sp);
    else
        clip.extent.y = GetYLimit(sp);

    if (hrMode==PalmOS5HR)
        HRRectangle(&clip);

    WinSetClip (&clip);
}

static void
CardDisplayStack (CardStackIndex stack)
{
    CardStackType   *sp = CardStackPtr (stack);
    CardType	    *cp;
    PatI	    pos;
    PatI	    avail;
    PatI	    off, face_off, pack_off, back_off;
    PatI	    base;
    CardIndex	    card;
    Card	    suit;
    RectangleType   oldClip;
    int		    count, face;

    sp->dirty = False;
    if (sp->display != CardDisplayNone)
    {
        SetStackClip (sp, &oldClip);

	CardEraseStack (stack);
 	CardSetShoulds (stack);

	if (sp->horizontal)
	{
	    sp->face_off = xOffset;
	    base = CoordXtoScreen(sp->pos.x);
	}
	else
	{
	    sp->face_off = yOffset;
            base = CoordYtoScreen(sp->pos.y);
	}
        if (game.alwaysPack)
            sp->pack_off = sp->face_off * 3 / 4;
        else
            sp->pack_off = sp->face_off;
	sp->back_off = cardBackOffset;

	avail = GetYLimit(sp) - base;

	/*
	 * Try to get the cards to fit on the screen
	 */
	/*
	 * First see if they all fit with the default spacing
	 */
	face_off = (sp->face_up + 1) * sp->face_off;
	pack_off = sp->pack_up * sp->pack_off;
	back_off = sp->back_up * sp->back_off;
	off = face_off + pack_off + back_off;
        if (off > avail)
	{
	    /*
	     * Second try to pack the 'packable' cards together
	     */
	    if (sp->pack_up)
		sp->pack_off = (avail - face_off - back_off) / sp->pack_up;
	    if (sp->pack_off < cardPackOffset)
		sp->pack_off = cardPackOffset;
	    pack_off = sp->pack_up * sp->pack_off;
	    off = face_off + pack_off + back_off;
	    if (off > avail)
	    {
		/*
		 * Third, set the pack spacing to a constant and
		 * try packing the face up cards a bit
		 */
		sp->pack_off = cardPackOffset;
		pack_off = sp->pack_up * sp->pack_off;
		sp->back_off = cardPackOffset;
		back_off = sp->back_up * sp->back_off;
		sp->face_off = (avail - pack_off - back_off) /
                        (sp->face_up + 1);
		/*
		 * If this would pack the face up cards closer together
		 * than the face down cards, space them all the same
		 */
		if (sp->face_off < sp->back_off)
		{
		    sp->face_off = (avail - pack_off) /
                            (sp->face_up + 1 + sp->back_up);
		    if (sp->face_off < sp->pack_off)
			sp->face_off = sp->pack_off;
		    sp->back_off = sp->face_off;
		}
		face_off = (sp->face_up + 1) * sp->face_off;
		back_off = sp->back_up * sp->back_off;
		off = face_off + pack_off + back_off;
		if (off > avail)
		{
		    /*
		     * Fourth, set the pack spacing to 0 and use
		     * all available space for the remaining cards
		     */
		    sp->pack_off = 0;
		    sp->back_off = cardPackOffset;
		    back_off = sp->back_up * sp->back_off;
		    sp->face_off = (avail - back_off) / (sp->face_up + 1);
		    if (sp->face_off < sp->back_off)
		    {
			sp->face_off = avail / (sp->face_up + 1 + sp->back_up);
			sp->back_off = sp->face_off;
		    }
		}
	    }
	}
        // end spacing checks

	pos = 0;
	count = 0;
	face = 0;
	if (sp->first == NullIndex)
	{
	    CardDraw (stack, pos, sp->empty);
	    pos++;
	}
	else
	{
	    for (card = sp->first; card != NullIndex; card = cp->next)
	    {
		cp = CardPtr (card);
		if (cp->face)
		{
		    suit = Suit(cp->card);
		    face++;
		}
		else
		    suit = BACK;
		if (cp->show != CardShowNone)
		{
		    CardDraw (stack, pos, MakeCard (0, suit, Rank (cp->card)));
		    pos++;
		}
		count++;
	    }
	}

	if (sp->display == CardDisplayTopCount && count)
	{
	    char	    buf[3];
	    char	    *b;

	    b = buf;
	    if (count >= 100)
	    {
		*b++ = '0' + count / 100;
		buf[0] = 
		count = count % 100;
	    }
	    if (count >= 10)
	    {
		*b++ = '0' + count / 10;
		count = count % 10;
	    }
	    *b++ = '0' + count;
            DrawCharsOnCard(buf, b-buf, CoordXtoScreen(sp->pos.x),
                    CoordYtoScreen(sp->pos.y), face==0, true);
	}
	WinSetClip (&oldClip);
    }
}

static void
DrawCharsOnCard(Char *buf, int len, Coord x, Coord y, Boolean centered,
        Boolean filled)
{
    int		    w, h;
    RectangleType   r;
    FontID          prevFont = -1;

    if (centered)
        prevFont = FntSetFont(boldFont);
    w = FntLineWidth (buf, len);
    h = FntLineHeight ();
    if (centered)
    {
        if (hrMode==PalmOS5HR)
        {
            x = x*2 + cardWidth - w / 2;
            y = y*2 + cardHeight - h / 2;
        }
        else
        {
            x += (cardWidth - w) / 2;
            y += (cardHeight - h) / 2;
        }
    }
    else
    {
        if (hrMode==PalmOS5HR)
        {
            x = x*2 + cardWidth*2 - w - 4 - cardSpaceX;
            y = y*2 + cardHeight*2 - h - cardSpaceY;
        }
        else
        {
            x += cardWidth - w - 2 - (cardSpaceX / 2);
            y += cardHeight - h - (cardSpaceY / 2);
        }
    }
    if (filled || romVersion < sysVersion35)
    {
        WinSetPattern (&emptyPattern);
        SetRectangle(&r, x-1, y, w+2, h);
        WinFillRectangle (&r, 0);
        WinDrawChars (buf, len, x, y);
    }
    else
    {
        WinPushDrawState ();
        WinSetDrawMode (winOverlay);
        WinPaintChars (buf, len, x, y);
        WinPopDrawState ();
    }
    if (centered)
        FntSetFont(prevFont);
}

#define HIT(_r,_x,_y)  (_r.topLeft.x <= _x && _x < _r.topLeft.x+_r.extent.x && \
			_r.topLeft.y <= _y && _y < _r.topLeft.y+_r.extent.y)

CardBool
CardHit (PatI u_x, PatI u_y, 
	 CardStackIndex *stackp, CardIndex *cardp,
	 PatI *c_x, PatI *c_y)
{
    CardStackType   *sp;
    CardType	    *cp;
    PatI    	    pos;
    CardStackIndex  stack;
    CardIndex	    card;
    RectangleType   r;
    PatI	    x, y, s_x, s_y;

    for (stack = 0; stack < game.nstacks; stack++)
    {
	sp = CardStackPtr (stack);
	if (sp->display != CardDisplayNone)
	{
	    pos = 0;
	    StackExtent (stack, &r);
	    if (HIT (r, u_x, u_y))
	    {
		s_x = r.topLeft.x;
		s_y = r.topLeft.y;
		*stackp = stack;
		*cardp = NullIndex;
		pos = 0;
		for (card = sp->first; card != NullIndex; card = cp->next)
		{
		    cp = CardPtr (card);
		    if (cp->show != CardShowNone)
		    {
			x = s_x;
			y = s_y;
			if (sp->horizontal)
			    x += StackOff (stack, pos, NULL);
			else
			    y += StackOff (stack, pos, NULL);
			CardExtent (x, y, &r);
			if (HIT (r, u_x, u_y))
			{
			    if (c_x)
				*c_x = x + cardWidth / 2;
			    if (c_y)
				*c_y = y + cardHeight / 2;
			    *cardp = card;
			}
			pos++;
		    }
		}
		return True;
	    }
	}
    }
    return False;
}

static PatI
CardPos (CardStackIndex stack, CardIndex card)
{
    CardStackType   *sp;
    PatI	    pos;
    CardType	    *cp;
    CardIndex	    c;

    sp = CardStackPtr (stack);
    if (sp->display == CardDisplayNone)
	return -1;
    pos = 0;
    for (c = sp->first; c != NullIndex; c = cp->next)
    {
	cp = CardPtr (c);
	if (cp->show != CardShowNone)
	{
	    if (c == card)
		return pos;
	    pos++;
	}
    }
    return -1;
}

static void
CardDisplayQuery (void)
{
    RectangleType   r;

    SetRectangle(&r, form_width-4, form_height-4, 4, 4);
    if (hrMode==PalmOS5HR)
        HRRectangle (&r);

    if (game.queryMode)
    {
	WinSetPattern (&myBlackPattern);
        WinFillRectangle (&r, 0);
    }
    else
        DrawTableBackground (&r);
    game.queryDirty = False;
}

static void
CardDisplayAlways (void)
{
    RectangleType   r;

    SetRectangle(&r, 0, form_height-4, 4, 4);
    if (hrMode==PalmOS5HR)
        HRRectangle (&r);

    if (game.alwaysAuto)
    {
	WinSetPattern (&myBlackPattern);
        WinFillRectangle (&r, 0);
    }
    else
        DrawTableBackground (&r);
    game.alwaysDirty = False;
}

void
CardRedisplay (void)
{
    CardStackIndex  stack;

    // Undo any hints before any drawing
    if (game.nhint > 0)
        CardUnshowHints ();

    if (romVersion >= sysVersion35)
    {
        WinPushDrawState ();
        WinSetForeColor (blackIndex);
        WinSetBackColor (whiteIndex);
        if (hrMode==PalmOS5HR)
            WinSetCoordinateSystem (kCoordinatesDouble);
        //WinScreenLock (winLockCopy);
    }

    if (game.queryDirty)
	CardDisplayQuery ();
    if (game.alwaysDirty)
	CardDisplayAlways ();
    for (stack = 0; stack < game.nstacks; stack++)
	if (CardStackPtr (stack)->dirty)
	    CardDisplayStack (stack);

    if (romVersion >= sysVersion35)
    {
        //WinScreenUnlock();
        WinPopDrawState ();
    }
}


// For highlighting -- saves the image while it is redrawn in a different
// style to highlight.  Should never need more than 5 (in montatana, when
// you click on an empty starting spot, it highlights all the 2s).
WinHandle hBuffer[MAX_HINT];
RectangleType hRect[MAX_HINT];
UInt8 hIndex = 0;

void
CardDrawHighlight (CardStackIndex stack, PatI pos)
{
    CardStackType   *sp = CardStackPtr (stack);
    CardType	    *cp = NULL;
    CardIndex	    card;
    Card	    suit;
    RectangleType   r, oldClip;
    PatI            tmp_pos, x, y;
    CardState       cardState;
    UInt16          error;

    x = CoordXtoScreen(sp->pos.x);
    y = CoordYtoScreen(sp->pos.y);
    if (sp->horizontal)
        x += StackOff (stack, pos, &cardState);
    else
        y += StackOff (stack, pos, &cardState);

    if (romVersion >= sysVersion35 && screen_depth > 1)
    {
        WinPushDrawState ();
        if (hrMode==PalmOS5HR)
            WinSetCoordinateSystem (kCoordinatesDouble);
        /* pick highlight colors and draw again */
        WinSetBackColor(hintIndex);

        SetStackClip (sp, &oldClip);

        // set card given pos
        tmp_pos = pos+1;
        for (card = sp->first; tmp_pos && card != NullIndex; card = cp->next)
        {
            cp = CardPtr (card);
            if (cp->show != CardShowNone)
                tmp_pos--;
        }

        // save current window appearance
        if (hIndex < MAX_HINT)
        {
            CardExtent(x, y, &r);
            hBuffer[hIndex] = CreateWindow(cardWidth, cardHeight, &error);
            CopyFromScreen(hBuffer[hIndex], &r, 0, 0);
            hRect[hIndex] = r;
            hIndex++;
        }
        //TODO else warning/error?

        // draw the card
        if (sp->first == NullIndex)
            CardDrawPos (x, y, GetXLimit(sp), GetYLimit(sp), CardBlank,
                    sp->horizontal, cardState);
        else
        {
            suit = (cp->face)? Suit(cp->card): BACK;
            CardDrawPos (x, y, GetXLimit(sp), GetYLimit(sp),
                    MakeCard (0, suit, Rank (cp->card)),
                    sp->horizontal, cardState);
        }

	WinSetClip (&oldClip);
        WinSetBackColor (whiteIndex);
        WinPopDrawState ();
    }
    else
    {
        CardRect (x, y, &r);
        WinInvertRectangle (&r, cardRadius);
    }
}

void
CardUndrawHighlight (CardStackIndex stack, PatI pos)
{
    RectangleType   r, oldClip;
    PatI            old_x=0, old_y=0;
    WinHandle       saveWin = NULL;

    if (romVersion >= sysVersion35 && screen_depth > 1)
    {
        ErrFatalDisplayIf (hIndex == 0, "hIndex == 0");
        if (hrMode==PalmOS5HR) {
            WinPushDrawState ();
            WinSetCoordinateSystem (kCoordinatesDouble);
        }
        SetStackClip (CardStackPtr (stack), &oldClip);
        // restore from the saved buffers
        hIndex--;
        saveWin = hBuffer[hIndex];
        old_x = hRect[hIndex].topLeft.x;
        old_y = hRect[hIndex].topLeft.y;
        CardExtent(0, 0, &r);
        if (saveWin)
        {
            CopyToScreen (saveWin, &r, old_x, old_y);
            WinDeleteWindow (saveWin, false);
        }
	WinSetClip (&oldClip);
        if (hrMode==PalmOS5HR)
            WinPopDrawState ();
    }
    else
        CardDrawHighlight (stack, pos);
}

WinHandle   animateBuffer, saveBuffer, cardBuffer;
BitmapType *cardBMP;
PatS	    animateDx, animateDy;
PatS	    animateX, animateY;
PatS	    animateOX, animateOY;
Card	    animateCard;
CardBool    animateHorizontal;

static void
CardSaveAnimate (void)
{
    RectangleType   r;

    CardExtent(animateX, animateY, &r);
    CopyFromScreen (saveBuffer, &r, 0, 0);
}

static void
CardPutAnimate (void)
{
    RectangleType   r;

    CardExtent(0, 0, &r);
    CopyToScreen (cardBuffer, &r, animateX, animateY);
}

static void
CardRestoreAnimate (void)
{
    RectangleType   r;

    CardExtent(0, 0, &r);
    CopyToScreen (saveBuffer, &r, animateOX, animateOY);
}

static void
CardDrawAnimate (void)
{
    PatI	    x, y, xcard, ycard, xsave, ysave;
    Boolean	    simple;

    simple = false;
    if (animateX > animateOX)
    {
	x = animateOX;
	xsave = 0;
	xcard = animateX - animateOX;
	if (xcard > cardWidth)
	    simple = true;
    }
    else
    {
	x = animateX;
	xsave = animateOX - animateX;
	xcard = 0;
	if (xsave > cardWidth)
	    simple = true;
    }
    if (animateY > animateOY)
    {
	y = animateOY;
	ysave = 0;
	ycard = animateY - animateOY;
	if (ycard > cardHeight)
	    simple = true;
    }
    else
    {
	y = animateY;
	ysave = animateOY - animateY;
	ycard = 0;
	if (ysave > cardHeight)
	    simple = true;
    }
    if (hrMode==PalmOS5HR) {
        WinPushDrawState ();
        WinSetCoordinateSystem (kCoordinatesDouble);
    }
    if (simple)
    {
	CardRestoreAnimate ();
	CardSaveAnimate ();
	CardPutAnimate ();
    }
    else
    {
        CardDrawAnimate2(x, y, xcard, ycard, xsave, ysave);
    }
    if (hrMode==PalmOS5HR)
        WinPopDrawState ();
}

static void
CardDrawAnimate2 (PatI x, PatI y, PatI xcard, PatI ycard,
        PatI xsave, PatI ysave)
{
    RectangleType   r;

    /*
     * Get bits from screen
     */
    SetRectangle (&r, x, y, cardWidth*2, cardHeight*2);
    CopyFromScreen (animateBuffer, &r, 0, 0);

    /*
     * Merge in saved bits
     */
    CardExtent (0, 0, &r);
    CopyRectangle (saveBuffer, animateBuffer, &r, xsave, ysave);

    /*
     * Create new saved bits
     */
    CardExtent (xcard, ycard, &r);
    CopyRectangle (animateBuffer, saveBuffer, &r, 0, 0);

    /*
     * Add in card image
     */
    CardExtent (0, 0, &r);
    if (romVersion >= sysVersion50) {
        // TODO use bitmap approach for 5.0 too
        // overlay on copy only works on 5.0 and later
        CopyRectangle2 (cardBuffer, animateBuffer, &r, xcard, ycard,
                winOverlay);
    } else if (romVersion >= sysVersion35) {
        // PalmOS 3.5 and 4.0 support transparency in bitmaps
        WinHandle draw = WinGetDrawWindow ();
        WinSetDrawWindow (animateBuffer);
        if (hrMode==SonyHR)
            HRWinDrawBitmap (hrRefNum, cardBMP, xcard*2, ycard*2);
        else
            WinDrawBitmap (cardBMP, xcard, ycard);
        WinSetDrawWindow (draw);
    }
    else
        CopyRectangle (cardBuffer, animateBuffer, &r, xcard, ycard);

    /*
     * stick the result on the screen
     */
    SetRectangle (&r, 0, 0, cardWidth*2, cardHeight*2);
    CopyToScreen (animateBuffer, &r, x, y);
}

CardBool
CardStartAnimate (PatI x, PatI y,
		  CardStackIndex stack, 
		  CardIndex card)
{
    UInt16	    error;
    CardStackType   *sp = CardStackPtr (stack);
    CardType	    *cp = CardPtr (card);
    Card	    suit;
    PatI	    pos;
    WinHandle	    draw;

    if (animateBuffer)
	CardStopAnimate ();
    pos = CardPos (stack, card);
    if (pos == -1)
	return False;

    animateBuffer = CreateWindow (cardWidth * 2, cardHeight * 2, &error);
    if (!animateBuffer)
	return False;

    saveBuffer = CreateWindow (cardWidth, cardHeight, &error);
    if (!saveBuffer)
    {
	WinDeleteWindow (animateBuffer, false);
	animateBuffer = 0;
	return False;
    }

    // 3.5 and 4.0 only support transparency in bitmaps, so draw
    // the card as a bitmap
    if (romVersion >= sysVersion35 && romVersion < sysVersion50)
        cardBuffer = CreateBitmapWindow (cardWidth, cardHeight, screen_depth,
                &cardBMP);
    else
        cardBuffer = CreateWindow (cardWidth, cardHeight, &error);

    if (!cardBuffer)
    {
        WinDeleteWindow (animateBuffer, false);
        animateBuffer = 0;
        WinDeleteWindow (saveBuffer, false);
        saveBuffer = 0;
        if (cardBMP)
            BmpDelete (cardBMP);
        cardBMP = NULL;
        return False;
    }

    animateDx = -cardWidth / 2;
    animateDy = -cardHeight / 2;
    animateX = x + animateDx;
    animateY = y + animateDy;
    animateHorizontal = sp->horizontal;
    if (cp->face)
	suit = Suit(cp->card);
    else
	suit = BACK;
    animateCard = MakeCard (0, suit, Rank (cp->card));
    draw = WinGetDrawWindow ();
    WinSetDrawWindow (cardBuffer);
    if (romVersion >= sysVersion35)
    {
        WinPushDrawState ();
        WinSetForeColor(blackIndex);
        WinSetBackColor(whiteIndex);
        if (hrMode==PalmOS5HR)
            WinSetCoordinateSystem (kCoordinatesDouble);
    }
    CardDrawPos (0, 0, GetXLimit(sp), GetYLimit(sp), animateCard,
            animateHorizontal, cardStateTop);
    WinSetDrawWindow (draw);
    CardSaveAnimate ();
    CardPutAnimate ();
    if (romVersion >= sysVersion35)
        WinPopDrawState ();
    return True;
}

void
CardStopAnimate (void)
{
    if (animateBuffer)
    {
	animateOX = animateX;
	animateOY = animateY;

        if (hrMode==PalmOS5HR) {
            WinPushDrawState ();
            WinSetCoordinateSystem (kCoordinatesDouble);
        }

	CardRestoreAnimate ();

        if (hrMode==PalmOS5HR)
            WinPopDrawState ();

	WinDeleteWindow (animateBuffer, false);
	WinDeleteWindow (saveBuffer, false);
	WinDeleteWindow (cardBuffer, false);
        if (cardBMP)
            BmpDelete (cardBMP);

	animateBuffer = 0;
	saveBuffer = 0;
	cardBuffer = 0;
        cardBMP = NULL;
    }
}


void
CardMoveAnimate (PatI x, PatI y)
{
    if (animateBuffer)
    {
	// [NJ] This change prevents animation when the movement is less
	// than 2 pixels.  That way, there is no quivering effect when the
	// user is trying to hold the stylus still.
	PatI newX, newY, dX, dY;

	newX = x + animateDx;
	newY = y + animateDy;
	dX = newX - animateX;
	dY = newY - animateY;
        //TODO pdc - x^2+y^2 seems like overkill for two pixels :-)
	if (dX * dX + dY * dY > 3) {
	    animateOX = animateX;
	    animateOY = animateY;
	    animateX = newX;
	    animateY = newY;
	    CardDrawAnimate ();
	}
    }
}
