/*
 * $Id: picture.c,v 1.7 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"

#define PictureDeck    0
#define PictureTable   1
#define PictureNtable  8
#define PictureLtable  (PictureTable + PictureNtable - 1)
#define PicturePile    (PictureTable + PictureNtable)
#define PictureNpile   24
#define PictureLpile   (PicturePile + PictureNpile - 1)
#define PictureAces    (PicturePile + PictureNpile)
#define PictureNull    (PictureAces + 1)
#define PictureNstack  (PictureNull + 1)

CardSetupType	PictureSetup[] = {
    { { 8, 4 }, { 8, 4 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartDown },

    { { 0, 3 }, { 0, 3 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 1, 3 }, { 1, 3 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 2, 3 }, { 2, 3 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 3, 3 }, { 3, 3 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 4, 3 }, { 4, 3 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 5, 3 }, { 5, 3 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 6, 3 }, { 6, 3 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 7, 3 }, { 7, 3 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },

    { { 0, 0 }, { 0, 0 }, CardEmpty, CardDisplayAll, False, False, 0, CardStartUp },
    { { 1, 0 }, { 1, 0 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 2, 0 }, { 2, 0 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 3, 0 }, { 3, 0 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 4, 0 }, { 4, 0 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 5, 0 }, { 5, 0 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 6, 0 }, { 6, 0 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 7, 0 }, { 7, 0 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },

    { { 0, 1 }, { 0, 1 }, CardEmpty, CardDisplayAll, False, False, 0, CardStartUp },
    { { 1, 1 }, { 1, 1 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 2, 1 }, { 2, 1 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 3, 1 }, { 3, 1 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 4, 1 }, { 4, 1 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 5, 1 }, { 5, 1 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 6, 1 }, { 6, 1 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 7, 1 }, { 7, 1 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },

    { { 0, 2 }, { 0, 2 }, CardEmpty, CardDisplayAll, False, False, 0, CardStartUp },
    { { 1, 2 }, { 1, 2 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 2, 2 }, { 2, 2 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 3, 2 }, { 3, 2 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 4, 2 }, { 4, 2 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 5, 2 }, { 5, 2 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 6, 2 }, { 6, 2 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },
    { { 7, 2 }, { 7, 2 }, CardEmpty, CardDisplayAll, False, False, 1, CardStartUp },

    { { 8, 0 }, { 8, 0 }, CardEmpty, CardDisplayAll, False, False, 0, CardStartUp },

    { { 8, 0 }, { 0, 0 }, CardEmpty, CardDisplayNone, False, False, 0, CardStartDown },
};

static void PictureInitialize (void) GAME_SECT;
static CardProgressType PictureCheckPile (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static void PictureMovePile (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType PictureCheckAces (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType PictureCheckEmptyTable (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static void PictureMoveDeal (CardStackIndex src, CardStackIndex dst,
        CardIndex last) GAME_SECT;
static CardBool PictureNormalize (void) GAME_SECT;

static void
PictureInitialize (void)
{
    CardStackIndex  stack, pile;
    CardStackType   *sp, *pilep;
    CardIndex	    card;
    CardType	    *cp;
    Card	    rank;
    Boolean         found;

    sp = CardStackPtr (PictureDeck);
    for (stack = PicturePile, rank = 4; stack <= PictureLpile; stack+=8, rank--)
    {
        found = False;
	for (card = sp->last; card != NullIndex; card = cp->prev)
	{
	    cp = CardPtr (card);
	    if (Rank (cp->card) == rank)
	    {
                CardTurn (PictureDeck, card, True, False);
		CardMoveCard (PictureDeck, card, card, stack, NullIndex, False);
                found = True;
		break;
	    }
	}
        if (!found)
        {
            // If we can't find a 2,3,4 in the deck, steal one from the table
            // and replace it with one from the deck
            for (pile = PicturePile; pile <= PictureLpile; pile++)
            {
                pilep = CardStackPtr (pile);
                if (Rank (CardPtr(pilep->last)->card) == rank)
                {
                    CardMove (pile, pilep->last, stack, False);
                    CardMove (PictureDeck, CardStackPtr (PictureDeck)->last,
                            pile, False);
                    CardTurn (pile, CardStackPtr (pile)->last, True, False);
                    break;
                }
            }
        }
    }

    for (stack = PicturePile; stack <= PictureLpile; stack++)
    {
        sp = CardStackPtr (stack);
        sp->empty = MakeCard(0, EMPTY, 4 - (stack-PicturePile)/8);
    }
}

static CardProgressType
PictureCheckPile (CardStackIndex	src,
	       CardStackIndex	dst,
	       CardIndex	card)
{
    Card	    rankbase;
    CardType        *cardp;
    CardType        *lastp;
    CardStackType   *sp = CardStackPtr (src);
    CardStackType   *dp = CardStackPtr (dst);

    if (card == NullIndex)
        return CardProgressCant;
    cardp = CardPtr (card);

    if (! cardp->face)
	return CardProgressCant;

    rankbase = 4 - (dst - PicturePile) / 8;

    // if same row (pile to pile) and already properly placed, don't allow
    if (src >= PicturePile && src <= PictureLpile &&
            ((src - PicturePile)/8) == ((dst - PicturePile)/8) &&
            Rank (CardPtr (sp->first)->card) == rankbase)
        return CardProgressCant;

    if (dp->last == NullIndex)
    {
	if (Rank (cardp->card) == rankbase)
	    return CardProgressSome;
    }
    else
    {
        if (Rank (CardPtr(dp->first)->card) != rankbase)
	    return CardProgressCant;
        lastp = CardPtr (dp->last);
	if (Suit(cardp->card) == Suit(lastp->card) &&
                Rank (cardp->card) == Rank (lastp->card) + 3)
	    return CardProgressSome;
    }
    return CardProgressCant;
}

static void
PictureMovePile (CardStackIndex	src,
	       CardStackIndex	dst,
	       CardIndex	card)
{
    CardIndex dlast = CardStackPtr (dst)->last;

    if (dlast != NullIndex)
        CardTurn (dst, dlast, False, True);
    CardMove (src, card, dst, True);
    // Move extra cards out of the deck so only two are visible
    if (dlast != NullIndex && dlast != CardStackPtr(dst)->first)
    {
        CardMoveCard (dst, dlast, dlast, PictureNull, NullIndex, True);
    }
}

static CardProgressType
PictureCheckAces (CardStackIndex	src,
	       CardStackIndex	dst,
	       CardIndex	card)
{
    if (Rank (CardPtr (card)->card) == 1)
        return CardProgressSome;
    return CardProgressCant;
}

static CardProgressType
    PictureCheckEmptyTable (CardStackIndex	src,
	       CardStackIndex	dst,
	       CardIndex	card)
{
    if (CardStackPtr(PictureDeck)->last != NullIndex)
        return CardProgressCant;
    if (CardStackPtr(dst)->last != NullIndex)
        return CardProgressCant;
    return CardProgressSome;
}

static void
PictureMoveDeal (CardStackIndex  src,
		CardStackIndex  dst,
		CardIndex	last)
{
    CardStackIndex  t;

    for (t = PictureTable; t <= PictureLtable; t++)
    {
	if (CardStackPtr (PictureDeck)->last == NullIndex)
	    return;
	CardMove (PictureDeck, CardStackPtr (PictureDeck)->last,
		  t, True);
	CardTurn (t, CardStackPtr (t)->last, True, True);
    }
}

CardMoveType	PictureMoves[] = {
    /* 0    Move card from pile to aces */
    {	"pile to aces",
	{ PicturePile, PictureLpile },
	{ PicturePile, PictureLpile },
	{ PictureAces, PictureAces },
	CardMoveAuto,
	0,
	PictureCheckAces,
	0
    },
    /* 1    Move card from table to aces */
    {	"table to aces",
	{ PictureTable, PictureLtable },
	{ PictureTable, PictureLtable },
	{ PictureAces, PictureAces },
	CardMoveAuto,
	0,
	PictureCheckAces,
	0
    },
    /* 2    Move card from pile to pile */
    {	"pile to pile",
	{ PicturePile, PictureLpile },
	{ PicturePile, PictureLpile },
	{ PicturePile, PictureLpile },
	CardMoveHelpful,
	0,
	PictureCheckPile,
	PictureMovePile
    },
    /* 3    Move card from table to pile */
    {	"table to pile",
	{ PictureTable, PictureLtable },
	{ PictureTable, PictureLtable },
	{ PicturePile, PictureLpile },
	CardMoveHelpful,
	0,
	PictureCheckPile,
	PictureMovePile
    },
    /* 4    Move card from table to empty table */
    {	"table to empty table",
	{ PictureTable, PictureLtable },
	{ PictureTable, PictureLtable },
	{ PictureTable, PictureLtable },
	CardMoveHelpful,
	0,
	PictureCheckEmptyTable,
        0
    },
    /* 5    deal */
    {	"deal",
	{ PictureDeck, PictureDeck },
	{ PictureDeck, PictureDeck },
	{ PictureTable, PictureLtable },
	CardMoveHelpful,
	0,
	0,
	PictureMoveDeal,
    },
};

#define NMOVES (sizeof (PictureMoves) / sizeof (PictureMoves[0]))

static CardBool
PictureNormalize (void)
{
    Card	    rankbase, ranktop;
    CardStackIndex  stack;
    CardStackType   *sp;
    
    for (stack = PicturePile; stack <= PictureLpile; stack++)
    {
        rankbase = 4 - (stack - PicturePile) / 8;
        ranktop = KING - (stack - PicturePile) / 8;
	sp = CardStackPtr (stack);
	if (sp->last == NullIndex)
	    return False;
        if (Rank (CardPtr (sp->first)->card) != rankbase ||
                Rank (CardPtr (sp->last)->card) != ranktop)
	    return False;
    }
    return True;
}

RulesType   pictureGame = {
    MakeCard (0, CLUB, 1), MakeCard (0, SPADE, KING), 2, PictureNstack,
    PictureNormalize,
    PictureInitialize,
    0,
    PictureSetup,
    NMOVES,
    PictureMoves
};
