/*
 * $Id: spider.c,v 1.7 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"
#include "spider.h"

CardSetupType	SpiderSetup[] =  {
    { { 0, 0 }, { 0, 0 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartDown },
    
    { { 2, 0 }, { 2, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 3, 0 }, { 3, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 4, 0 }, { 4, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 5, 0 }, { 5, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 6, 0 }, { 6, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 7, 0 }, { 7, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 8, 0 }, { 8, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    { { 9, 0 }, { 9, 0 }, CardEmpty, CardDisplayTop, False, False, 0, CardStartDown },
    
    { { 0, 1 }, { 0, 1 }, CardNone, CardDisplayPack, False, False, 6, CardStartLastUp },
    { { 1, 1 }, { 1, 1 }, CardNone, CardDisplayPack, False, False, 5, CardStartLastUp },
    { { 2, 1 }, { 2, 1 }, CardNone, CardDisplayPack, False, False, 5, CardStartLastUp },
    { { 3, 1 }, { 3, 1 }, CardNone, CardDisplayPack, False, False, 6, CardStartLastUp },
    { { 4, 1 }, { 4, 1 }, CardNone, CardDisplayPack, False, False, 5, CardStartLastUp },
    { { 5, 1 }, { 5, 1 }, CardNone, CardDisplayPack, False, False, 5, CardStartLastUp },
    { { 6, 1 }, { 6, 1 }, CardNone, CardDisplayPack, False, False, 6, CardStartLastUp },
    { { 7, 1 }, { 7, 1 }, CardNone, CardDisplayPack, False, False, 5, CardStartLastUp },
    { { 8, 1 }, { 8, 1 }, CardNone, CardDisplayPack, False, False, 5, CardStartLastUp },
    { { 9, 1 }, { 9, 1 }, CardNone, CardDisplayPack, False, False, 6, CardStartLastUp },
};


static void SpiderMoveDeal (CardStackIndex  src, CardStackIndex  dst, CardIndex last) GAME_SECT;
static CardBool SpiderNormalize (void) GAME_SECT;


CardBool
SpiderSelectTable (CardStackIndex   stack,
		   CardIndex	    *first,
		   CardIndex	    *last)
{
    CardStackType   *sp = CardStackPtr (stack);

    // select largest range in suit order
    *first = sp->last;
    *last = CardInReverseSuitOrder (sp->last);
    return True;
}

CardProgressType
SpiderCheckTable (CardStackIndex    src, 
		  CardStackIndex    dst,
		  CardIndex	    card)
{
    CardIndex	    prev;
    CardStackType   *dp = CardStackPtr (dst);
    CardIndex       tmpcard;
    CardType        *cp;
    unsigned int    count = 0;

    // card must be face-up and moved to a stack preserving rank order
    // (CheckEmpty is for moving to an empty stack)
    if (!CardPtr (card)->face || dp->last == NullIndex ||
            !CardIsInOrder (card, dp->last) ||
            CardIsInSuitOrder (card, dp->last))
	return CardProgressCant;

    // Hint rankings (dest not in suit):
    // source NOT in suit
    //    prefer highest dest rank, then shortest dest stack
    // source in suit
    //    no hint but allowed

    prev = CardPtr (card)->prev;
    if (prev != NullIndex)
    {
        // discourage move if the card is already in suit
        // (but if so, how did we get here?)
        if (CardIsInSuitOrder (card, prev))
            return CardProgressNoHint;

        if (CardIsInOrder (card, prev))
        {
            // discourage swaps between non-suited piles
            if (Rank (CardPtr (dp->last)->card) ==
                    Rank (CardPtr (prev)->card))
                return CardProgressNoHint;
            return CardProgressNone;
        }
    }

    //experimental-> return the rank of the card, which will prefer
    //the largest ranked card not in any order

    count++;    // count dest
    tmpcard = dp->last;
    while ((cp = CardPtr (tmpcard))->prev != NullIndex && 
	   CardIsInSuitOrder (tmpcard, cp->prev))
    {
        tmpcard = cp->prev;
        count++;
    }
    return CardProgressNone +
            Rank (CardPtr (dp->last)->card) * 13 +
            13 - count;
}

/*
 * Like CheckTable except requires card and dest to be in suit order
 */
CardProgressType
SpiderCheckSuitTable  (CardStackIndex	src, 
		       CardStackIndex	dst,
		       CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);
    CardIndex       tmpcard;
    CardType        *cp;
    unsigned int    count = 0;

    if (!CardPtr (card)->face || !CardIsInSuitOrder (card, dp->last))
	return CardProgressCant;

    // Wether source is already in an in-suit stack or not, prefer to make
    // the largest combined stack, breaking ties with lowest dest rank
    // (final tie breaker keep current stack?)

    count++;    // count dest
    tmpcard = dp->last;
    while ((cp = CardPtr (tmpcard))->prev != NullIndex && 
	   CardIsInSuitOrder (tmpcard, cp->prev))
    {
        tmpcard = cp->prev;
        count++;
    }

    count++;    // count 'card'
    tmpcard = card;
    while ((cp = CardPtr (tmpcard))->next != NullIndex && 
	   CardIsInSuitOrder (cp->next, tmpcard))
    {
        tmpcard = cp->next;
        count++;
    }
    // add 200 to be greater than CheckTable return values
    // (KING*13 + 13 = 182)
    return count + CardProgressNone + 200;
}

CardProgressType
SpiderCheckEmpty (CardStackIndex    src, 
		  CardStackIndex    dst,
		  CardIndex	    card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (!CardPtr (card)->face)
	return CardProgressCant;
    // destination is empty
    if (dp->last == NullIndex)
    {
        // discourage empty <-> empty swaps
	if (CardPtr (card)->prev == NullIndex)
	    return CardProgressNoHint;
	return CardProgressNone; // don't interfere with other "check" hints
    }
    return CardProgressCant;
}

CardProgressType
SpiderCheckPile (CardStackIndex src, 
		 CardStackIndex dst,
		 CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);
    CardStackType   *sp = CardStackPtr (src);

    if (!CardPtr (card)->face)
	return CardProgressCant;
    if (dp->last != NullIndex)
	return CardProgressCant;
    if (Rank (CardPtr (card)->card) != KING)
	return CardProgressCant;
    if (Rank (CardPtr (sp->last)->card) != 1)
	return CardProgressCant;
    return CardProgressSome;
}

static void
SpiderMoveDeal (CardStackIndex  src,
		CardStackIndex  dst,
		CardIndex	last)
{
    CardStackIndex  t;

    for (t = SpiderTable; t <= SpiderLtable; t++)
    {
	if (CardStackPtr (SpiderDeck)->last == NullIndex)
	    return;
	CardMove (SpiderDeck, CardStackPtr (SpiderDeck)->last,
		  t, True);
	CardTurn (t, CardStackPtr (t)->last, True, True);
    }
}

CardMoveType SpiderMoves[] = {
    // 2.6.2 - For some reason, this was fourth, which didn't work,
    //         since table to table is a subset of a complete build.
    //         It seems to work just fine first.
    /* 3    Move card from table to pile */
    {	"table to pile",
	{ SpiderTable, SpiderLtable },
	{ SpiderTable, SpiderLtable },
	{ SpiderPile, SpiderLpile },
	CardMoveAuto,
	SpiderSelectTable,
	SpiderCheckPile,
	0,
    },
    /* 0    Move card from table to table same suit */
    // actually makes for different behavior depending on whether the move
    // is from a stack in order or not, which is strange
    {	"table to table in suit",
	{ SpiderTable, SpiderLtable },
	{ SpiderTable, SpiderLtable },
	{ SpiderTable, SpiderLtable },
	CardMoveHelpful,
	SpiderSelectTable,
	SpiderCheckSuitTable,
	0,
    },
    /* 1    Move card from table to table */
    {	"table to table",
	{ SpiderTable, SpiderLtable },
	{ SpiderTable, SpiderLtable },
	{ SpiderTable, SpiderLtable },
	CardMoveHelpful,
	SpiderSelectTable,
	SpiderCheckTable,
	0,
    },
    /* 2    Move card from table to empty */
    {	"table to empty",
	{ SpiderTable, SpiderLtable },
	{ SpiderTable, SpiderLtable },
	{ SpiderTable, SpiderLtable },
	CardMoveHelpful,
	SpiderSelectTable,
	SpiderCheckEmpty,
	0,
    },
    /* 4    deal */
    {	"deal",
	{ SpiderDeck, SpiderDeck },
	{ SpiderDeck, SpiderDeck },
	{ SpiderTable, SpiderLtable },
	CardMoveHelpful,
	0,
	0,
	SpiderMoveDeal,
    },
};

#define NMOVES	(sizeof (SpiderMoves) / sizeof (SpiderMoves[0]))

static CardBool
SpiderNormalize (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    CardIndex	    card;
    CardType	    *cp;

    // Turn up any face down cards revealed by playing other cards
    for (stack = SpiderTable; stack <= SpiderLtable; stack++)
    {
	sp = CardStackPtr (stack);
	card = sp->last;
	if (card != NullIndex)
	{
	    cp = CardPtr (card);
	    if (!cp->face)
		CardTurn (stack, card, True, True);
	}
    }

    for (stack = SpiderPile; stack <= SpiderLpile; stack++)
    {
	sp = CardStackPtr (stack);
	card = sp->last;
	if (card == NullIndex)
	    return False;
    }
    return True;
}

RulesType   spiderGame = {
    MakeCard (0, CLUB, 1), MakeCard (0, SPADE, KING), 2, SpiderNstack,
    SpiderNormalize,
    0,
    CardIsInSuitOrder,
    SpiderSetup,
    NMOVES,
    SpiderMoves
};

// Same as spider, but with spades and hearts only
RulesType   tarantulaGame = {
    MakeCard (0, HEART, 1), MakeCard (0, SPADE, KING), 4, SpiderNstack,
    SpiderNormalize,
    0,
    CardIsInSuitOrder,
    SpiderSetup,
    NMOVES,
    SpiderMoves
};

// too easy!
#if 0
// Same as spider, but with spades only (with only one suit, in effect,
// builds don't have to be in build order)
RulesType   blackWidowGame = {
    MakeCard (0, SPADE, 1), MakeCard (0, SPADE, KING), 8, SpiderNstack,
    SpiderNormalize,
    0,
    CardIsInSuitOrder,
    SpiderSetup,
    NMOVES,
    SpiderMoves
};
#endif
