/*
 * $Id: stalactites.c,v 1.7 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"

#define StalactitesDeck 	0
#define StalactitesTemp 	1
#define StalactitesNtemp	2
#define StalactitesLtemp	(StalactitesTemp + StalactitesNtemp - 1)
#define StalactitesPile 	(StalactitesLtemp + 1)
#define StalactitesNpile	4
#define StalactitesLpile	(StalactitesPile + StalactitesNpile - 1)
#define StalactitesTable	(StalactitesLpile + 1)
#define StalactitesNtable	8
#define StalactitesLtable	(StalactitesTable + StalactitesNtable - 1)
#define StalactitesNstack	(StalactitesLtable + 1)

CardSetupType	StalactitesSetup[] =  {
    { { 0, 0 }, { 0, 0 }, CardEmpty, CardDisplayNone, False, False, 0, CardStartDown },

    { { 0, 0 }, { 0, 0 }, CardBlank, CardDisplayTop, False, False, 0, CardStartUp },
    { { 1, 0 }, { 1, 0 }, CardBlank, CardDisplayTop, False, False, 0, CardStartUp },

    { { 3, 0 }, { 3, 0 }, CardEmpty, CardDisplayTopCount, False, False, 1, CardStartUp },
    { { 4, 0 }, { 4, 0 }, CardEmpty, CardDisplayTopCount, False, False, 1, CardStartUp },
    { { 5, 0 }, { 5, 0 }, CardEmpty, CardDisplayTopCount, False, False, 1, CardStartUp },
    { { 6, 0 }, { 6, 0 }, CardEmpty, CardDisplayTopCount, False, False, 1, CardStartUp },

    { { 0, 1 }, { 0, 1 }, CardNone, CardDisplayAll, False, True, 6, CardStartUp },
    { { 1, 1 }, { 1, 1 }, CardNone, CardDisplayAll, False, True, 6, CardStartUp },
    { { 2, 1 }, { 2, 1 }, CardNone, CardDisplayAll, False, True, 6, CardStartUp },
    { { 3, 1 }, { 3, 1 }, CardNone, CardDisplayAll, False, True, 6, CardStartUp },
    { { 4, 1 }, { 4, 1 }, CardNone, CardDisplayAll, False, True, 6, CardStartUp },
    { { 5, 1 }, { 5, 1 }, CardNone, CardDisplayAll, False, True, 6, CardStartUp },
    { { 6, 1 }, { 6, 1 }, CardNone, CardDisplayAll, False, True, 6, CardStartUp },
    { { 7, 1 }, { 7, 1 }, CardNone, CardDisplayAll, False, True, 6, CardStartUp },
};

static CardProgressType StalactitesCheckPile (CardStackIndex src,
        CardStackIndex dst, CardIndex	card) GAME_SECT;
static CardProgressType StalactitesCheckTemp (CardStackIndex src,
        CardStackIndex dst, CardIndex	card) GAME_SECT;
static CardBool StalactitesNormalize (void) GAME_SECT;

static CardProgressType
StalactitesCheckPile (CardStackIndex src, 
		 CardStackIndex dst,
		 CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);
    CardType        *firstp, *cardp;

    firstp = CardPtr(dp->first);
    cardp = CardPtr(card);

    if (Rank(cardp->card) != Rank(firstp->card) &&
            CardIsInCircularOrder (dp->last, card))
        return CardProgressSome;
    return CardProgressCant;
}

static CardProgressType
StalactitesCheckTemp (CardStackIndex src, 
		 CardStackIndex dst,
		 CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (dp->last == NullIndex)
	return CardProgressSome;
    return CardProgressCant;
}

CardMoveType StalactitesMoves[] = {
    /* 0    Move card from table to pile */
    {	"table to pile",
	{ StalactitesTable, StalactitesLtable },
	{ StalactitesTable, StalactitesLtable },
	{ StalactitesPile, StalactitesLpile },
	CardMoveAuto,
	0,
	StalactitesCheckPile,
	0,
    },
    /* 1    Move card from temp to pile */
    {	"temp to pile",
	{ StalactitesTemp, StalactitesLtemp },
	{ StalactitesTemp, StalactitesLtemp },
	{ StalactitesPile, StalactitesLpile },
	CardMoveAuto,
	0,
	StalactitesCheckPile,
	0,
    },
    /* 2    Move card from table to temp */
    {	"table to temp",
	{ StalactitesTable, StalactitesLtable },
	{ StalactitesTable, StalactitesLtable },
	{ StalactitesTemp, StalactitesLtemp },
	CardMoveHelpful,
	0,
	StalactitesCheckTemp,
	0,
    },
};

#define NMOVES	(sizeof (StalactitesMoves) / sizeof (StalactitesMoves[0]))

static CardBool
StalactitesNormalize (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;

    for (stack = StalactitesPile; stack <= StalactitesLpile; stack++)
    {
	sp = CardStackPtr (stack);
        if (CardIsInCircularOrder (sp->last, sp->first))
            continue;
        else
            return False;
    }
    return True;
}

RulesType   stalactitesGame = {
    MakeCard (0, CLUB, 1), MakeCard (0, SPADE, KING), 1, StalactitesNstack,
    StalactitesNormalize,
    0,
    CardIsInSuitOrder,
    StalactitesSetup,
    NMOVES,
    StalactitesMoves
};
