/*
 * $Id: stats.c,v 1.10 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Some code derived from from the ListView code from the Memopad samples.
 *
 * Portions copyright (c) 2000 Palm, Inc. or its subsidiaries.
 * All rights reserved.
 */

// needed for numRows hack, PalmOS 3.3 and earlier
#define ALLOW_ACCESS_TO_INTERNALS_OF_TABLES
#define ALLOW_ACCESS_TO_INTERNALS_OF_FIELDS

#include "patience.h"
#include "patienceRsc.h"
#include <System/SysEvtMgr.h>
#include <SonyCLIE.h>
#include <WinGlue.h>
// HandEra
#include <Vga.h>
#include <stdio.h>


/*
 * Change log
 *
 * $Log: stats.c,v $
 * Revision 1.10  2005/01/11 17:19:00  pdclose
 * Update copyright notice
 *
 * Revision 1.9  2005/01/10 01:22:27  pdclose
 * - Implement custom drawing code so we can format the table the way we like
 *   (left-justified bold game names, non-bold numbers).
 * - Move the scrollbar to the left if the left-handed preference is set.
 * - Support PalmOS Dynamic Input Area (see patience.c).
 * - Use an ugly hack to make the table look good in PalmOS 3.3 and earlier,
 *   yet still allow resizing in modern units.
 * - Implement a slightly more efficient table redraw algorithm when scrolling.
 * 
 * Revision 1.8  2004/04/06 15:59:20  pdclose
 * Copyright date update
 * 
 * Revision 1.7  2004/04/05 22:38:14  pdclose
 * Merge form sizing between stats and new forms
 * 
 * Revision 1.6  2004/03/25 02:35:46  pdclose
 * Fix table height logic
 * Move stats to union in preferences
 * 
 * Revision 1.5  2003/03/03 23:22:00  pdclose
 * New games, resize routines for Sony as well as HandEra.
 * 
 * Revision 1.4  2002/09/19 19:26:53  pdclose
 * New scheme for game names.
 * Try to merge new game list/table with stats display.  In progress.
 * 
 * Revision 1.3  2002/07/06 21:10:27  pdclose
 * Add new games.
 * 
 * Revision 1.2  2002/06/17 01:28:37  pdclose
 * Add new games: Black Widow, Tarantula, and Forty Thieves.
 * 
 * Revision 1.1  2002/04/23 21:54:34  pdclose
 * Initial revision
 */

static Boolean	StatsFormHandleEvent(EventPtr) CARD_SECT;
static void	StatsScroll (FormPtr, TablePtr, ScrollBarPtr, Int16) CARD_SECT;
static UInt16	RowsInTable (TablePtr) CARD_SECT;
static void	StatsLoadTable (FormPtr, TablePtr, ScrollBarPtr) CARD_SECT;
static void	StatsLoadRow (TablePtr, UInt16) CARD_SECT;
static void     StatsDrawRecord (void *, Int16, Int16, RectanglePtr) CARD_SECT;
static void     NewGameDrawRecord (void *, Int16, Int16, RectanglePtr)
    CARD_SECT;
static Int16	GameCompare (void *, void *, Int32) CARD_SECT;
static void	SizeStatsForm (FormPtr) CARD_SECT;
static void *	GetObjectPtrByType (FormPtr, FormObjectKind) CARD_SECT;

static Int16 row_offset = 0;

// Same order as 'games' in patience.c
enum
{
    gameKlondike=0, gameMontana, gameTowers, gameSpider, gameTabby,
    gameGolf, gameAces, gameCalc, gameYukon, gameSpiderette, gameEightOff,
    gameCanfield, gameVegas, gameFreecell, gameWish, gameStalactites,
    gameTarantula, gameFortyThieves, gamePictureGallery, gameMissMilligan,
    gameRoyalCotillion, gameTerrace, gameBakersDozen, gameGypsy
};

static Char game_name[MAX_GAMES][24];
static UInt8 game_sort[MAX_GAMES];
static enum { currentStatsForm, currentNewGameForm } current_form;

static Int16
GameCompare (void *A, void *B, Int32 other)
{
    return StrCompare (game_name[*((UInt8*)A)], game_name[*((UInt8*)B)]);
}

void
StatisticsForm (void)
{
    UInt16	btn;
    UInt16	answer;
    TablePtr    table;
    UInt16	row;
    UInt16      rows;
    FormPtr	statsFrm = FrmInitForm (statsForm);

    if (statsFrm)
    {
        SizeStatsForm (statsFrm);

        current_form = currentStatsForm;

        for (row=0; row < num_games; row++)
        {
            game_sort[row] = row;
            SysCopyStringResource (game_name[row], nameFirst + row);
#if 0
            // TODO Should resize table columns based on max width of labels
            // TODO Could use STRINGTABLE and SysFormPointerArrayToStrings
            nameH = DmGetResource ('tSTR', id);
            if (nameH)
            {
                name = MemHandleLock(nameH);
                // len(name)
                // new(len)
                // copy
                MemHandleUnlock (nameH);
                DmReleaseResource (nameH);
            }
            else
                game_name[row] = "**Not Initialized**";
#endif
        }

        SysInsertionSort (game_sort, num_games, sizeof(UInt8), GameCompare, 0);

        table = GetObjectPtrById(statsFrm, statsTable);

        TblHasScrollBar (table, true);

        TblSetColumnUsable (table, 0, true);
        TblSetColumnUsable (table, 1, true);
        TblSetColumnUsable (table, 2, true);
        TblSetColumnUsable (table, 3, true);
	TblSetCustomDrawProcedure (table, 0, StatsDrawRecord);
	TblSetCustomDrawProcedure (table, 1, StatsDrawRecord);
	TblSetCustomDrawProcedure (table, 2, StatsDrawRecord);
	TblSetCustomDrawProcedure (table, 3, StatsDrawRecord);

        rows = TblGetNumberOfRows (table);
        for ( ; row < rows; row++)
        {
            TblSetRowUsable(table, row, false);
            //TblMarkRowInvalid (table, row);
            TblSetRowSelectable (table, row, false);
        }
        // BUG Palm OS 3.3 and earlier draw boxes at the bottom of the table
        if (romVersion < sysVersion35)
        {
            // This is a hack, but I can't find a way around it.  The table
            // size has to be defined as larger than this for the devices
            // that can hide the silkscreen.
            table->numRows = 10;
        }

        StatsLoadTable(statsFrm, table,
                GetObjectPtrById(statsFrm, statsScrollBar));

        for (;;)
        {
            btn = DoDialog (statsFrm, StatsFormHandleEvent);
            switch (btn)
            {
            case statsDone:
                // no action
                break;
            case statsClear:
                // are you sure?
                answer = FrmAlert (statsClearAlert);
                if (answer == 0)
                {
                    // clear stats
                    MemSet(&game.u.stats, sizeof(game.u.stats), 0);
                }
                break;
            default:
                break;
            }
            break;
        }
	FrmDeleteForm (statsFrm);
    }
}

void
NewGameForm (Boolean entertain)
{
    UInt16	btn;
    FormPtr	gameFrm = FrmInitForm (newGameForm);
    TablePtr    table;
    UInt16	row;
    UInt16	column;
    UInt16      rows;
    Int16	gameKind = -1;
    UInt16	prevGameKind = game.gameKind;
    ControlPtr	cancelBtn;

    // TODO Implement categories to filter games
    // (one deck, two decks, perhaps most played)

    // TODO merge common code with StatsForm
    if (gameFrm)
    {
        SizeStatsForm (gameFrm);

        current_form = currentNewGameForm;

        for (row=0; row < num_games; row++)
        {
            game_sort[row] = row;
            SysCopyStringResource (game_name[row], nameFirst + row);
#if 0
            // TODO Should resize table columns based on max width of labels
            // Could use STRINGTABLE and SysFormPointerArrayToStrings
            nameH = DmGetResource ('tSTR', id);
            if (nameH)
            {
                name = MemHandleLock(nameH);
                // len(name)
                // new(len)
                // copy
                MemHandleUnlock (nameH);
                DmReleaseResource (nameH);
            }
            else
                game_name[row] = "**Not Initialized**";
#endif
        }

        SysInsertionSort (game_sort, num_games, sizeof(UInt8), GameCompare, 0);

        table = GetObjectPtrById(gameFrm, newGameTable);

        TblHasScrollBar (table, true);

        TblSetColumnUsable (table, 0, true);
	TblSetCustomDrawProcedure (table, 0, NewGameDrawRecord);

        rows = TblGetNumberOfRows (table);
        for ( ; row < rows; row++)
        {
            TblSetRowUsable(table, row, false);
            //TblMarkRowInvalid (table, row);
            TblSetRowSelectable (table, row, false);
        }
        // BUG Palm OS 3.3 and earlier draw boxes at the bottom of the table
        if (romVersion < sysVersion35)
        {
            // This is a hack, but I can't find a way around it.  The table
            // size has to be defined as larger than this for the devices
            // that can hide the silkscreen.
            table->numRows = 10;
        }

        StatsLoadTable(gameFrm, table,
                GetObjectPtrById(gameFrm, newGameScrollBar));

        // TODO pre-select the current game

        /*
         * Only enable the cancel button if a game is in progress
         */
        cancelBtn = GetObjectPtrById (gameFrm, newGameCancelBtn);
        CtlSetUsable (cancelBtn, game.playing);
        
        if (entertain)
            DialogIdleFunc = SelectGameEntertain;

        for (;;)
        {
            btn = DoDialog (gameFrm, StatsFormHandleEvent);
            switch (btn)
            {
            case newGameNewBtn:
                if ( TblGetSelection (table, &row, &column) ) 
                    gameKind = game_sort[row + row_offset];
                    // TODO Default to something if no game was selected?
                    //      But still handle cancel...
                break;
            case newGameRulesBtn:
                if ( TblGetSelection (table, &row, &column) ) 
                    gameKind = game_sort[row + row_offset];
                if (0 <= gameKind && gameKind < num_games)
                    Rules (rulesFirst + gameKind);
                continue;
            case newGameCancelBtn:
                gameKind = -1;
                break;
            default:
                break;
            }
            break;
        }
	FrmDeleteForm (gameFrm);
    }

    if (0 <= gameKind && gameKind < num_games)
    {
        // record stats for previous game
        if (prevGameKind >= 0 && prevGameKind < num_games)
        {
            if (game.playing)
                game.u.stats[prevGameKind].lost++;
            else
                game.u.stats[prevGameKind].won++;
        }
	game.gameKind = gameKind;
	game.gameSeed = TimGetTicks ();
	game.playing = True;
	game.queryMode = False;
	RestartGame ();
    }
}

static Boolean StatsFormHandleEvent(EventPtr event)
{
    FormPtr             frm = NULL;
    TablePtr 		table = NULL;
    ScrollBarPtr 	scrollbar = NULL;
    UInt16		rows;
    Boolean             handled = false;

    switch (event->eType) {
    case frmUpdateEvent:
        if (window_resized)
        {
            frm = FrmGetActiveForm ();
            SizeStatsForm (frm);
            table = GetObjectPtrByType (frm, frmTableObj);
            scrollbar = GetObjectPtrByType (frm, frmScrollBarObj);
            if (table == NULL || scrollbar == NULL)
                return false;
            StatsLoadTable (frm, table, scrollbar);
            TblRedrawTable (table);
            window_resized = false;
        }
        // don't handle to allow system redraw
        break;
    // PalmOS dynamic input area
    case winEnterEvent:
        if (diaMode) {
            // PIM 1.0 compatility
            PINSetInputTriggerState(pinInputTriggerEnabled);
            PINSetInputAreaState(pinInputAreaUser);
        }
        // don't think we need to handle this, we didn't really do anything
        break;
    // PalmOS dynamic input area
    case winDisplayChangedEvent:
        window_resized = true;
        // not needed?
        //FrmUpdateForm (FrmGetActiveFormID (), frmRedrawUpdateCode);
        handled = true;
        break;
    // HandEra
    case displayExtentChangedEvent:
        window_resized = true;
        FrmUpdateForm (FrmGetActiveFormID (), frmRedrawUpdateCode);
        handled = true;
        break;
    case sclRepeatEvent:
        frm = FrmGetActiveForm ();
        table = GetObjectPtrByType (frm, frmTableObj);
        scrollbar = GetObjectPtrByType (frm, frmScrollBarObj);
        if (table == NULL || scrollbar == NULL)
            return false;
        StatsScroll(frm, table, scrollbar,
                event->data.sclRepeat.newValue - event->data.sclRepeat.value);
        //handled = true;
	break;
    case keyDownEvent:
        frm = FrmGetActiveForm ();
        table = GetObjectPtrByType (frm, frmTableObj);
        scrollbar = GetObjectPtrByType (frm, frmScrollBarObj);
        if (table == NULL || scrollbar == NULL)
            return false;
        rows = RowsInTable (table);
	switch (event->data.keyDown.chr) {
	case pageDownChr:
	    StatsScroll (frm, table, scrollbar, rows-1);
	    handled = true;
	    break;
	case pageUpChr:
	    StatsScroll (frm, table, scrollbar, -rows+1);
	    handled = true;
	    break;
	}
        SclSetScrollBar (GetObjectPtrByType (frm, frmScrollBarObj),
                row_offset, 0, num_games-rows, rows);
	break;
    default:
    }
    return handled;
}

static void
SizeStatsForm (FormPtr form)
{
    UInt16          formObjIndex;
    Int16           heightDelta;
    Int16           tableHeight;
    Int16           lineDelta;
    RectangleType   r;

    if (qvgaMode || silkMode || diaMode)
    {
        SizeFormWindow (form, NULL, &heightDelta);

        // if table increased in size, reset row_offset to avoid viewing
        // items off the end of the table
        if (heightDelta > 0)
            row_offset = 0;

        // done/clear -  adjust y
        formObjIndex = FrmGetObjectIndex (form, statsDone);
        if (formObjIndex == frmInvalidObjectId)
            formObjIndex = FrmGetObjectIndex (form, newGameNewBtn);
        FrmGetObjectBounds (form, formObjIndex, &r);
        r.topLeft.y += heightDelta;
        tableHeight = r.topLeft.y; // just a starting point
        FrmSetObjectBounds (form, formObjIndex, &r);

        AdjustObject (form, statsClear, 0, heightDelta, 0, 0);
        AdjustObject (form, newGameRulesBtn, 0, heightDelta, 0, 0);
        AdjustObject (form, newGameCancelBtn, 0, heightDelta, 0, 0);

        // table/scrollbar adjust height
        formObjIndex = FrmGetObjectIndex (form, statsTable);
        if (formObjIndex == frmInvalidObjectId)
            formObjIndex = FrmGetObjectIndex (form, newGameTable);
        FrmGetObjectBounds (form, formObjIndex, &r);
        // calculate table height based on position of "done" button,
        // otherwise the rounding accumulates with multiple minimize /
        // maximize actions
        tableHeight -= r.topLeft.y;
        // round down so all rows fit evenly on the screen
        lineDelta = (tableHeight % FntLineHeight ());
        r.extent.y = tableHeight - lineDelta;
        FrmSetObjectBounds (form, formObjIndex, &r);

        formObjIndex = FrmGetObjectIndex (form, statsScrollBar);
        if (formObjIndex == frmInvalidObjectId)
            formObjIndex = FrmGetObjectIndex (form, newGameScrollBar);
        FrmGetObjectBounds (form, formObjIndex, &r);
        r.extent.y = tableHeight - lineDelta;
        FrmSetObjectBounds (form, formObjIndex, &r);
    }

    // Move scrollbar to left for left-handed layout

    if (game.leftHanded)
    {
        UInt16          scrollObjIndex;
        RectangleType   scrollRect;
        UInt16          tableObjIndex;
        RectangleType   tableRect;
        Int16           tableDelta;

        scrollObjIndex = FrmGetObjectIndex (form, statsScrollBar);
        if (scrollObjIndex == frmInvalidObjectId)
            scrollObjIndex = FrmGetObjectIndex (form, newGameScrollBar);
        FrmGetObjectBounds (form, scrollObjIndex, &scrollRect);

        tableObjIndex = FrmGetObjectIndex (form, statsTable);
        if (tableObjIndex == frmInvalidObjectId)
            tableObjIndex = FrmGetObjectIndex (form, newGameTable);
        FrmGetObjectBounds (form, tableObjIndex, &tableRect);

        if (scrollRect.topLeft.x > tableRect.topLeft.x)
        {
            // delta = scroll x + width - (table x + width)
            tableDelta = scrollRect.topLeft.x + scrollRect.extent.x -
                    (tableRect.topLeft.x + tableRect.extent.x);
            scrollRect.topLeft.x = tableRect.topLeft.x;
            tableRect.topLeft.x += tableDelta;

            FrmSetObjectBounds (form, scrollObjIndex, &scrollRect);
            FrmSetObjectBounds (form, tableObjIndex, &tableRect);

            AdjustObject (form, statsLabelGame, tableDelta, 0, 0, 0);
            AdjustObject (form, statsLabelWins, tableDelta, 0, 0, 0);
            AdjustObject (form, statsLabelLosses, tableDelta, 0, 0, 0);
            AdjustObject (form, statsLabelPct, tableDelta, 0, 0, 0);
            AdjustObject (form, newGameLabelGame, tableDelta, 0, 0, 0);
            AdjustObject (form, newGameLabelWins, tableDelta, 0, 0, 0);
            AdjustObject (form, newGameLabelLosses, tableDelta, 0, 0, 0);
        }
    }
}

static UInt16
RowsInTable (TablePtr table)
{
    UInt16          rows;
    UInt16          rows_in_table;
    //FontID	    currFont;
    RectangleType   r;

    rows_in_table = TblGetNumberOfRows (table);

    // calculate number of rows for current font
    TblGetBounds (table, &r);
    //currFont = FntSetFont (boldFont);
    rows = r.extent.y / FntLineHeight ();
    //FntSetFont (currFont);

    if (rows <= rows_in_table)
        return rows;
    else
        return rows_in_table;
}

static void
StatsScroll (FormPtr frm, TablePtr table, ScrollBarPtr scrollbar,
        Int16 linesToScroll)
{
    Int16 		i;
    UInt16		rows;
    UInt16		lastRow;
    UInt16 		scrollAmount;
    RectangleType	scrollR;
    RectangleType	vacated;
    WinDirectionType	direction;

    if (frm == NULL || table == NULL || scrollbar == NULL)
        return;

    // Move the bits that will remain visible.

    rows = RowsInTable (table);

    if (linesToScroll > 0)
    {
        if (linesToScroll >= rows)
            linesToScroll = rows-1;
        if (row_offset + rows + linesToScroll > num_games)
            linesToScroll = num_games - row_offset - rows;
    }
    else
    {
        if (-linesToScroll >= rows)
            linesToScroll = -rows+1;
        if (row_offset + linesToScroll <= 0)
            linesToScroll = -row_offset;
    }

    if (((linesToScroll > 0) && (linesToScroll < rows)) ||
            ((linesToScroll < 0) && (-linesToScroll < rows)))
    {
        scrollAmount = 0;

        if (linesToScroll > 0)
        {
            lastRow = TblGetLastUsableRow (table) - 1;
            for (i = 0; i < linesToScroll; i++)
            {
                scrollAmount += TblGetRowHeight (table, lastRow);
                TblRemoveRow (table, 0);
                // remove scrolls the table contents up, so we always want
                // to load the last one
                StatsLoadRow (table, rows-1);
            }
            direction = winUp;
        }
        else
        {
            for (i = 0; i < -linesToScroll; i++)
            {
                scrollAmount += TblGetRowHeight (table, 0);
                TblInsertRow (table, 0);
                // insert scrolls the table contents down, so we always want
                // to load the first one
                StatsLoadRow (table, 0);
            }
            direction = winDown;
        }

        TblGetBounds (table, &scrollR);
        WinScrollRectangle (&scrollR, direction, scrollAmount, &vacated);
        WinEraseRectangle (&vacated, 0);
    }

    row_offset += linesToScroll;

    // Update the scrollbar
    SclSetScrollBar (scrollbar, row_offset, 0, num_games-rows, rows);

    TblRedrawTable (table);
}

static void
StatsLoadTable (FormPtr frm, TablePtr table, ScrollBarPtr scrollbar)
{
    UInt16		row;
    UInt16		rows;

    rows = RowsInTable (table);
    for (row = 0; row < rows; row++)
        StatsLoadRow (table, row);

    // Update the scrollbar
    SclSetScrollBar (scrollbar, row_offset, 0, num_games-rows, rows);
}

static void
StatsLoadRow (TablePtr table, UInt16 row)
{
    TblSetRowUsable (table, row, true);
    TblSetRowSelectable (table, row, false);
    // Mark the row invalid so that it will draw when we
    // call the draw routine.
    TblMarkRowInvalid (table, row);
    //currFont = FntSetFont (boldFont);
    TblSetRowHeight (table, row, FntLineHeight ());
    //FntSetFont (currFont);

    if (current_form == currentStatsForm)
    {
        // Use custom so we control spacing, fonts
        TblSetItemStyle (table, row, 0, customTableItem);
        TblSetItemStyle (table, row, 1, customTableItem);
        TblSetItemStyle (table, row, 2, customTableItem);
        TblSetItemStyle (table, row, 3, customTableItem);
    }
    else // newGameForm
    {
        TblSetItemStyle (table, row, 0, customTableItem);
    }
}

static void
StatsDrawRecord (void *table, Int16 row, Int16 column, RectanglePtr bounds)
{
    Int16   x, y;
    UInt16  len, pos, width;
    Char    posStr[7];
    UInt16  gameIndex = game_sort[row + row_offset];
    Char *  gameName = game_name[gameIndex];
    PatUL wins = game.u.stats[gameIndex].won;
    PatUL losses = game.u.stats[gameIndex].lost;
    FontID  currFont;

#if 0
    // column sizing test
    wins = gameIndex*1000;
    losses = (23-gameIndex)*1000;
#endif
#if 0
    // font-sizing debug
    FntSetFont (stdFont);
    wins = FntLineWidth(gameName, StrLen(gameName)));
    FntSetFont (boldFont);
    losses = FntLineWidth(gameName, StrLen(gameName)));
    FntSetFont (stdFont);
#endif

    // column widths (std bold large lgbold):
    //  Picture Gallery 61  77  77  76
    //  Royal Cotillion 57  73  74  72
    //  Baker's Dozen   57  70  74  71
    //  Forty Thieves   57  69  72  71
    //  Miss Milligan   51  65  67  65
    //  single digit     5   6   7   6
    //  %                8  13  11  11
    //  4 digits        20  24  28  24
    //  5 digits        25  30  35  30
    //  xxx%            23  31  32  29
    //  xx%             18  25  25  23
    //      (100% is only possible if losses are 0, so
    //       there's room for the 1 in the losses column)

    //x = bounds->topLeft.x + 1; ???
    x = bounds->topLeft.x;
    y = bounds->topLeft.y;
    width = bounds->extent.x;

    switch (column)
    {
      case 0:   // game name
        currFont = FntSetFont (boldFont);
        len = StrLen (gameName);
        // palmos 3.1 (or can use glue)
        WinGlueDrawTruncChars(gameName, len, x, y, width);
        FntSetFont (currFont);
        break;
      case 1:   // wins
        StrIToA (posStr, wins);
        len = StrLen(posStr);
        pos = ( width - FntCharsWidth (posStr, len) ) / 2;
        WinDrawChars (posStr, len, x + pos, y);
        break;
      case 2:   // losses
        StrIToA (posStr, losses);
        len = StrLen(posStr);
        pos = ( width - FntCharsWidth (posStr, len) ) / 2;
        WinDrawChars (posStr, len, x + pos, y);
        break;
      case 3:   // percent
        if (wins + losses > 0) {
            StrIToA (posStr, wins * 100 / (wins + losses));
            StrCat (posStr, "%");
        } else
            StrCopy (posStr, "----");
        len = StrLen(posStr);
        pos = width - FntCharsWidth (posStr, len);
        WinDrawChars (posStr, len, x + pos, y);
        break;
    }
}

static void
NewGameDrawRecord (void *table, Int16 row, Int16 column, RectanglePtr bounds)
{
    Int16   x, y;
    Char    posStr[7];
    UInt16  len, pos, width;
    UInt16  gameIndex = game_sort[row + row_offset];
    Char *  gameName = game_name[gameIndex];
    PatUL   wins = game.u.stats[gameIndex].won;
    PatUL   losses = game.u.stats[gameIndex].lost;
    FontID  currFont;

    x = bounds->topLeft.x + 1;
    y = bounds->topLeft.y;

    currFont = FntSetFont (boldFont); // could also be largeFont, largeBoldFont

    // Game name
    width = 78;
    len = StrLen (gameName);
    WinGlueDrawTruncChars(gameName, len, x, y, width);

    // Wins
    x += width;
    width = 31;

    StrIToA (posStr, wins);
    len = StrLen(posStr);
    pos = ( width - FntCharsWidth (posStr, len) ) / 2;
    WinDrawChars (posStr, len, x + pos, y);

    // Losses
    x += width;
    width = 31;

    StrIToA (posStr, losses);
    len = StrLen(posStr);
    pos = ( width - FntCharsWidth (posStr, len) ) / 2;
    WinDrawChars (posStr, len, x + pos, y);

    FntSetFont (currFont);
}

static void *
GetObjectPtrByType (FormPtr frm, FormObjectKind obj_type)
{
    int obj_index, num_objs;

    num_objs = FrmGetNumberOfObjects (frm);
    for (obj_index = 0; obj_index < num_objs; obj_index++)
    {
        if (FrmGetObjectType (frm, obj_index) == obj_type)
        {
            return FrmGetObjectPtr (frm, obj_index);
        }
    }
    return NULL;
}
